#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import copy
from typing import Any, Dict, Optional

import numpy as np
import torch
import torch.nn as nn
from flsim.active_user_selectors.simple_user_selector import (
    SequentialActiveUserSelectorConfig,
)
from flsim.clients.base_client import ClientConfig
from flsim.clients.dp_client import DPClientConfig
from flsim.common.pytest_helper import assertEqual, assertEmpty, assertNotEmpty
from flsim.optimizers.async_aggregators import FedAvgWithLRHybridAggregatorConfig
from flsim.optimizers.local_optimizers import LocalOptimizerSGDConfig
from flsim.privacy.common import PrivacySetting
from flsim.reducers.base_round_reducer import ReductionType
from flsim.reducers.weighted_dp_round_reducer import WeightedDPRoundReducerConfig
from flsim.servers.aggregator import AggregationType
from flsim.servers.sync_dp_servers import SyncDPSGDServerConfig
from flsim.servers.sync_servers import SyncServerConfig
from flsim.tests.utils import (
    MetricsReporterWithMockedChannels,
    FakeMetricReporter,
    verify_models_equivalent_after_training,
)
from flsim.trainers.sync_trainer import SyncTrainer, SyncTrainerConfig
from flsim.utils.async_trainer.async_staleness_weights import (
    PolynomialStalenessWeightConfig,
)
from flsim.utils.fl.common import FLModelParamUtils
from flsim.utils.sample_model import DummyAlphabetFLModel
from flsim.utils.tests.helpers.async_trainer_test_utils import (
    get_fl_data_provider,
    run_fl_training,
)
from flsim.utils.tests.helpers.test_data_utils import DummyAlphabetDataset
from omegaconf import OmegaConf
from opacus import PrivacyEngine


class TestDifferentialPrivacyIntegration:
    def _get_weighted_dp_reducer_config(
        self,
        noise,
        clipping,
        reduction_type=ReductionType.WEIGHTED_SUM,
        min_weight=1e-6,
        max_weight=1.0,
    ):
        return WeightedDPRoundReducerConfig(
            reduction_type=reduction_type,
            min_weight=min_weight,
            max_weight=max_weight,
            privacy_setting=PrivacySetting(
                noise_multiplier=noise, clipping_value=clipping
            ),
        )

    def _create_optimizer(
        self,
        model: nn.Module,
        data_provider,
        lr: float,
        momentum: float,
        sample_level_dp: bool,
        dp_config: Dict[str, Any],
        train_batch_size: int,
    ):
        optimizer = torch.optim.SGD(model.parameters(), lr=lr, momentum=momentum)
        optimizer.zero_grad()
        if sample_level_dp:
            pe = PrivacyEngine(
                model,
                batch_size=train_batch_size,
                sample_size=self.data_size,  # pyre-ignore
                alphas=dp_config["alphas"],
                noise_multiplier=dp_config["sample_dp_noise_multiplier"],
                max_grad_norm=dp_config["sample_dp_clipping_value"],
            )
            pe.attach(optimizer)
        return optimizer

    def _load_data(self, one_user: bool, data_size: int = 26):
        """
        Loads the data, which is a Dummy alphabet, either for 1 user with
        `data_size` samples, or for N (`data_size`) users, each with 1 sample.
        """
        if one_user:  # the single user gets the whole shard (which is the data)
            shard_size = data_size
            local_batch_size = data_size
        else:  # will have N (`data_size`) users, each with one sample
            shard_size = 1
            local_batch_size = 1
        dummy_dataset = DummyAlphabetDataset(num_rows=data_size)
        (
            data_provider,
            data_loader,
        ) = DummyAlphabetDataset.create_data_provider_and_loader(
            dummy_dataset, shard_size, local_batch_size, DummyAlphabetFLModel()
        )
        assertEqual(data_loader.num_total_users, data_size / shard_size)
        assertEqual(data_loader.num_total_users, data_provider.num_users())
        self.data_size = data_size  # pyre-ignore
        return data_provider, data_loader.train_batch_size

    def _train_vanilla_pytorch_dp_model(
        self,
        lr: float,
        momentum: float,
        sample_level_dp: bool,
        dp_config: Dict[str, Any],
    ):
        """
        Trains a vanilla Pytorch DP model, without FL (there is 1 user)
        """
        model_wrapper = DummyAlphabetFLModel()
        data_provider, train_batch_size = self._load_data(one_user=True)

        optimizer = self._create_optimizer(
            model_wrapper.model,
            data_provider,
            lr,
            momentum,
            sample_level_dp,
            dp_config,
            train_batch_size,
        )
        model_wrapper.model.train()
        for one_user_data in data_provider.train_data():
            for batch in one_user_data:
                batch_metrics = model_wrapper.fl_forward(batch)
                loss = batch_metrics.loss
                loss.backward()
                optimizer.step()

        return model_wrapper

    def _train_fl_model(
        self,
        lr: float,
        momentum: float,
        one_user: bool,
        dp_config: Optional[Dict[str, Any]] = None,
        noise_func_seed: Optional[int] = None,
        data_size: int = 26,
        **kwargs,
    ):
        """
        Trains an FL model, with or without DP
        """
        # create dummy FL model on alphabet
        global_fl_model = DummyAlphabetFLModel()
        data_provider, _ = kwargs.pop(
            "data_provider", self._load_data(one_user, data_size)
        )
        world_size = 1

        metrics_reporter = FakeMetricReporter()
        if one_user:
            users_per_round = 1
        else:  # otherwise, we have `data_size` users
            users_per_round = data_size
        epochs = kwargs.pop("epoch", 1)
        users_per_round = kwargs.pop("users_per_round", users_per_round)
        metrics_reporter = kwargs.pop("metrics_reporter", metrics_reporter)
        eval_epoch_frequency = kwargs.pop("eval_epoch_frequency", 1.0)
        train_metrics_reported_per_epoch = kwargs.pop(
            "train_metrics_reported_per_epoch", 1
        )
        aggregation_type = kwargs.pop("aggregation_type", AggregationType.AVERAGE)

        sync_trainer = SyncTrainer(
            model=global_fl_model,
            cuda_enabled=False,
            **OmegaConf.structured(
                SyncTrainerConfig(
                    users_per_round=users_per_round,
                    epochs=epochs,
                    always_keep_trained_model=False,
                    train_metrics_reported_per_epoch=train_metrics_reported_per_epoch,
                    report_train_metrics=True,
                    eval_epoch_frequency=eval_epoch_frequency,
                    do_eval=True,
                    report_train_metrics_after_aggregation=True,
                    client=DPClientConfig(
                        epochs=1,
                        optimizer=LocalOptimizerSGDConfig(lr=lr, momentum=momentum),
                        privacy_setting=PrivacySetting(
                            alphas=dp_config["alphas"],
                            noise_multiplier=dp_config["sample_dp_noise_multiplier"],
                            clipping_value=dp_config["sample_dp_clipping_value"],
                            target_delta=dp_config["delta"],
                            noise_seed=noise_func_seed,
                        ),
                    )
                    if dp_config is not None
                    else ClientConfig(
                        epochs=1,
                        optimizer=LocalOptimizerSGDConfig(lr=lr, momentum=momentum),
                    ),
                    server=SyncDPSGDServerConfig(
                        active_user_selector=SequentialActiveUserSelectorConfig(),
                        privacy_setting=PrivacySetting(
                            alphas=dp_config["alphas"],
                            noise_multiplier=dp_config["user_dp_noise_multiplier"],
                            clipping_value=dp_config["user_dp_clipping_value"],
                            target_delta=dp_config["delta"],
                            noise_seed=noise_func_seed,
                        ),
                        aggregation_type=aggregation_type,
                    )
                    if dp_config is not None
                    else SyncServerConfig(),
                )
            ),
        )

        global_fl_model, _eval_metric = sync_trainer.train(
            data_provider,
            metrics_reporter,
            num_total_users=data_provider.num_users(),
            distributed_world_size=world_size,
        )

        return global_fl_model

    def test_dp_turned_off_by_params(self) -> None:
        """
        Tests DP and no-DP produce the same exact model, when DP parameters are off.
        Basically, tests the equivalence of calling SyncServer and SyncDPServer with noise = 0 and clip = inf
        """
        lr = 0.1
        momentum = 0.0

        # first, call SyncTrainer (DP is off)
        torch.manual_seed(1)
        fl_model_with_vanilla_server = self._train_fl_model(
            lr=lr, momentum=momentum, one_user=False, dp_config=None
        )
        # set DP parameters off.
        off_dp_config = {
            "alphas": [10, 100],
            "sample_dp_noise_multiplier": 0.0,
            "sample_dp_clipping_value": float("inf"),
            "user_dp_noise_multiplier": 0.0,
            "user_dp_clipping_value": float("inf"),
            "delta": 0.00001,
        }
        torch.manual_seed(1)
        fl_model_with_dp_server = self._train_fl_model(
            lr=lr, momentum=momentum, one_user=False, dp_config=off_dp_config
        )
        error_msg = verify_models_equivalent_after_training(
            fl_model_with_vanilla_server,
            fl_model_with_dp_server,
            model_init=None,
            rel_epsilon=1e-6,
            abs_epsilon=1e-6,
        )
        assertEmpty(error_msg, msg=error_msg)

    def test_dp_ineffective(self) -> None:
        """
        Tests DP and no-DP produce the same exact model, when DP parameters are ineffective.
        Basically, tests the equivalence of the following 2 scenarios, with N users:

        1. Calling SyncTrainer (this is "no DP")
        2. Calling PrivateSyncTrainer, and DP (this is "DP") when parameters are ineffective

        Note:
            To make the dp_config ineffective, we need to set noise = 0 and clipping value to
            a large number. This is different from test test_dp_turned_off_by_params() that sets
            clipping value to inf, as this test actually applies DP, but because of parameter
            values dp is ineffective, the results would be identical to when DP is OFF.
        """
        lr = 0.1
        momentum = 0.0

        torch.manual_seed(1)
        # Call vanilla sync server
        fl_model_with_vanilla_server = self._train_fl_model(
            lr=lr, momentum=momentum, one_user=False, dp_config=None
        )

        # Next call set DP parameters ineffective.
        ineffective_dp_config = {
            "alphas": [10, 100],
            "sample_dp_noise_multiplier": 0.0,
            "sample_dp_clipping_value": 9999.9,
            "user_dp_noise_multiplier": 0.0,
            "user_dp_clipping_value": 9999.9,
            "delta": 0.00001,
        }
        torch.manual_seed(1)
        fl_model_with_dp_server = self._train_fl_model(
            lr=lr,
            momentum=momentum,
            one_user=False,
            dp_config=ineffective_dp_config,
        )

        error_msg = verify_models_equivalent_after_training(
            fl_model_with_vanilla_server,
            fl_model_with_dp_server,
            model_init=None,
            rel_epsilon=1e-6,
            abs_epsilon=1e-6,
        )
        assertEmpty(error_msg, msg=error_msg)

    def test_frameworks_one_client_sample_dp_off(self) -> None:
        """
        Tests if Pytorch-DP and FL simulator, generate the same resutls, when
        sample-level DP is off. Essentially, tests equivalence of the following
        scenarios, when there is 1 user:

        1. Using FL simulator, when sample-level DP is off
        2. Using Pytorch DP, when sample-level DP is off

        Note:
            For this test, user-level DP is off, since it is not relavant to Pytorch-DP's
            provided functionality, which is sample-level DP.
        """
        lr = 1e-2
        momentum = 0.9

        torch.manual_seed(1)
        fl_model = self._train_fl_model(lr=lr, momentum=momentum, one_user=True)
        torch.manual_seed(1)
        vanilla_dp_model = self._train_vanilla_pytorch_dp_model(
            lr=lr, momentum=momentum, sample_level_dp=False, dp_config={}
        )
        assertEqual(
            FLModelParamUtils.get_mismatched_param(
                [fl_model.fl_get_module(), vanilla_dp_model.fl_get_module()], 1e-6
            ),
            "",
        )

    def test_frameworks_one_client_clipping_only(self) -> None:
        """
        Tests if Pytorch-DP and FL simulator, generate the same resutls, when there
        is only 1 user and when sample-level DP is on (with noising off).
        Essentially, tests equivalence of the following scenarios:

        1. Using FL simulator, when there is one user and DP is on (with noise=0),
        2. Using Pytorch DP (one user), when DP is on,
        (DP parameters are the same for the 2 scenarios.)

        Note:
            For this test, user-level DP is off, since it is not relavant to Pytorch-DP's
            provided functionality, which is sample-level DP.
        """
        lr = 0.25
        momentum = 0.0
        alphas = [1 + x / 10.0 for x in range(1, 100)] + [
            float(y) for y in list(range(12, 64))
        ]
        dp_config = {
            "alphas": alphas,
            "sample_dp_noise_multiplier": 0.0,
            "sample_dp_clipping_value": 1.0,
            "user_dp_noise_multiplier": 0.0,
            "user_dp_clipping_value": float("inf"),
            "delta": 0.00001,
        }
        torch.manual_seed(1)
        fl_model = self._train_fl_model(
            lr=lr, momentum=momentum, one_user=True, dp_config=dp_config
        )
        torch.manual_seed(1)
        vanilla_dp_model = self._train_vanilla_pytorch_dp_model(
            lr=lr, momentum=momentum, sample_level_dp=True, dp_config=dp_config
        )
        assertEqual(
            FLModelParamUtils.get_mismatched_param(
                [fl_model.fl_get_module(), vanilla_dp_model.fl_get_module()], 1e-6
            ),
            "",
        )

    def test_user_dp_equivalent_sample_dp_when_one_client_one_example(
        self,
    ) -> None:
        """
        Tests if user-level DP is equivalent to sample-level DP, when we
        have one user, and that user has one example (one epoch)

        """
        lr = 1.0
        momentum = 0.0

        alphas = [1 + x / 10.0 for x in range(1, 100)] + [
            float(y) for y in list(range(12, 64))
        ]

        dp_config_user_dp_off = {
            "alphas": alphas,
            "sample_dp_noise_multiplier": 0.7,
            "sample_dp_clipping_value": 1.0,
            "user_dp_noise_multiplier": 0.0,
            "user_dp_clipping_value": float("inf"),
            "delta": 0.00001,
        }
        torch.manual_seed(1)
        dp_model_sample_level_dp_on = self._train_fl_model(
            lr=lr,
            momentum=momentum,
            one_user=True,
            dp_config=dp_config_user_dp_off,
            data_size=1,
            noise_func_seed=1234,
        )

        dp_config_sample_dp_off = {
            "alphas": alphas,
            "sample_dp_noise_multiplier": 0,
            "sample_dp_clipping_value": float("inf"),
            "user_dp_noise_multiplier": 0.7,
            "user_dp_clipping_value": 1.0,
            "delta": 0.00001,
        }
        torch.manual_seed(1)
        dp_model_user_level_dp_on = self._train_fl_model(
            lr=lr,
            momentum=momentum,
            one_user=True,
            dp_config=dp_config_sample_dp_off,
            data_size=1,
            noise_func_seed=1234,
        )

        error_msg = verify_models_equivalent_after_training(
            dp_model_sample_level_dp_on,
            dp_model_user_level_dp_on,
            model_init=None,
            rel_epsilon=1e-6,
            abs_epsilon=1e-6,
        )
        assertEmpty(error_msg, msg=error_msg)

    def test_user_dp_equivalent_sample_dp(self) -> None:
        """
        Tests if user-level DP is equivalent to sample-level DP under a certain
        degenerate condition. It tests the equivalence of the following:

        1. User-level DP, with N users, 1 example each (can be N different examples),
        2. Sample-level DP, with 1 user, N examples (can be N different examples),

        under these Conditions:

        Condition 1. [Just as a sanity check] when DP is off (user-level and sample-level)
        Condition 2. When DP is on, noise_multiplier for both cases is 0 (i.e. clipping only)
        Condition 2. When DP is on, noise_multiplier for both cases is 1, manual seed is set

        Note:
            For both cases, we set lr = 1.0 and momentum = 0.0

        """
        lr = 1.0
        momentum = 0.0

        # Condition 1
        torch.manual_seed(1)
        no_dp_model_one_user = self._train_fl_model(
            lr=lr, momentum=momentum, one_user=True
        )
        torch.manual_seed(1)
        no_dp_model = self._train_fl_model(lr=lr, momentum=momentum, one_user=False)

        error_msg = verify_models_equivalent_after_training(
            no_dp_model_one_user,
            no_dp_model,
            model_init=None,
            rel_epsilon=1e-6,
            abs_epsilon=1e-6,
        )
        assertEmpty(error_msg, msg=error_msg)

        # Condition 2
        alphas = [1 + x / 10.0 for x in range(1, 100)] + [
            float(y) for y in list(range(12, 64))
        ]
        dp_config_user_dp_off = {
            "alphas": alphas,
            "sample_dp_noise_multiplier": 0.0,
            "sample_dp_clipping_value": 0.8,
            "user_dp_noise_multiplier": 0,
            "user_dp_clipping_value": 0.8,
            "delta": 0.00001,
        }
        torch.manual_seed(1)
        dp_model_one_user_sample_dp = self._train_fl_model(
            lr=lr, momentum=momentum, one_user=True, dp_config=dp_config_user_dp_off
        )
        dp_config_sample_dp_off = {
            "alphas": alphas,
            "sample_dp_noise_multiplier": 0.0,
            "sample_dp_clipping_value": 0.8,
            "user_dp_noise_multiplier": 0.0,
            "user_dp_clipping_value": 0.8,
            "delta": 0.00001,
        }
        torch.manual_seed(1)
        dp_model_user_dp = self._train_fl_model(
            lr=lr,
            momentum=momentum,
            one_user=False,
            dp_config=dp_config_sample_dp_off,
        )

        error_msg = verify_models_equivalent_after_training(
            dp_model_one_user_sample_dp,
            dp_model_user_dp,
            model_init=None,
            rel_epsilon=1e-6,
            abs_epsilon=1e-6,
        )
        assertEmpty(error_msg, msg=error_msg)

        # Condition 3
        dp_config_user_dp_off = {
            "alphas": alphas,
            "sample_dp_noise_multiplier": 1,
            "sample_dp_clipping_value": 0.5,
            "user_dp_noise_multiplier": 0.0,
            "user_dp_clipping_value": 0.5,
            "delta": 0.00001,
        }
        torch.manual_seed(1)
        dp_model_one_user_sample_dp = self._train_fl_model(
            lr=lr,
            momentum=momentum,
            one_user=True,
            dp_config=dp_config_user_dp_off,
            noise_func_seed=1000,
        )
        dp_config_sample_dp_off = {
            "alphas": alphas,
            "sample_dp_noise_multiplier": 0.0,
            "sample_dp_clipping_value": 0.5,
            "user_dp_noise_multiplier": 1,
            "user_dp_clipping_value": 0.5,
            "delta": 0.00001,
        }
        torch.manual_seed(1)
        dp_model_user_dp = self._train_fl_model(
            lr=lr,
            momentum=momentum,
            one_user=False,
            dp_config=dp_config_sample_dp_off,
            noise_func_seed=1000,
        )
        error_msg = verify_models_equivalent_after_training(
            dp_model_one_user_sample_dp,
            dp_model_user_dp,
            None,
            rel_epsilon=1e-6,
            abs_epsilon=1e-6,
        )
        assertEmpty(error_msg, msg=error_msg)

    def test_private_trainer_reporting(self) -> None:
        lr = 1.0
        momentum = 0.0
        alphas = [1 + x / 10.0 for x in range(1, 100)] + [
            float(y) for y in list(range(12, 64))
        ]
        dp_config = {
            "alphas": alphas,
            "sample_dp_noise_multiplier": 0.0,
            "sample_dp_clipping_value": 2.0,
            "user_dp_noise_multiplier": 1.0,
            "user_dp_clipping_value": 2.0,
            "delta": 0.00001,
        }
        torch.manual_seed(1)
        metrics_reporter = MetricsReporterWithMockedChannels()
        self._train_fl_model(
            lr=lr,
            momentum=momentum,
            one_user=False,
            dp_config=dp_config,
            users_per_round=4,
            epochs=5,
            metrics_reporter=metrics_reporter,
            train_metrics_reported_per_epoch=10,
            eval_epoch_frequency=0.1,
        )
        # we have 26 users, 4 users per round, that makes 7 rounds per epcoh
        # we are reporting train, aggregation, and eval metrics all
        # at 10 rounds per epoch, so we should get a total of 21 reports

        def count_word(result, word):
            return str(result).count(word)

        # check for existance of sample_dp and user_dp in reports.
        # they are logged only in Aggrefation and we aggregated 7 times.

        # print to std our prints sample_dp wich is a dict of four values once
        # hence we should get 7 occurrences.
        assertEqual(
            count_word(metrics_reporter.stdout_results, "sample level dp"),
            7,
            metrics_reporter.stdout_results,
        )

        # summary writer breaks dict to 4 plots, hence we get 28 occurrences.
        assertEqual(
            count_word(metrics_reporter.tensorboard_results, "sample level dp"),
            28,
            metrics_reporter.tensorboard_results,
        )

        # for user_dp we only log one value per round.
        assertEqual(
            count_word(metrics_reporter.stdout_results, "user level dp"),
            7,
            metrics_reporter.stdout_results,
        )

        assertEqual(
            count_word(metrics_reporter.tensorboard_results, "user level dp"),
            7,
            metrics_reporter.tensorboard_results,
        )

    def _test_dp_no_dp_same_weighted_async(
        self,
        noise,
        clip_norm,
        data_provider,
        buffer_size,
        epochs=1,
    ):
        local_lr = np.random.sample()
        global_lr = np.random.sample()

        dp_model = DummyAlphabetFLModel()
        nondp_model = copy.deepcopy(dp_model)
        staleness_config = PolynomialStalenessWeightConfig(
            exponent=0.5, avg_staleness=0
        )

        aggregator_config_nondp = FedAvgWithLRHybridAggregatorConfig(
            lr=global_lr, buffer_size=buffer_size
        )
        aggregator_config_dp = FedAvgWithLRHybridAggregatorConfig(
            lr=global_lr, buffer_size=buffer_size
        )

        for reduction_type in [
            ReductionType.WEIGHTED_SUM,
            ReductionType.WEIGHTED_AVERAGE,
        ]:
            nondp_fl_trained_model, _ = run_fl_training(
                fl_model=nondp_model,
                fl_data_provider=data_provider,
                epochs=epochs,
                local_lr=local_lr,
                aggregator_config=aggregator_config_nondp,
                training_duration_mean=1,
                staleness_weight_config=staleness_config,
            )

            reducer_config = self._get_weighted_dp_reducer_config(
                noise=noise,
                clipping=clip_norm,
                reduction_type=reduction_type,
            )
            aggregator_config_dp.reducer = reducer_config
            dp_fl_trained_model, _ = run_fl_training(
                fl_model=dp_model,
                fl_data_provider=data_provider,
                epochs=epochs,
                local_lr=local_lr,
                aggregator_config=aggregator_config_dp,
                training_duration_mean=1,
                staleness_weight_config=staleness_config,
            )

            return verify_models_equivalent_after_training(
                nondp_fl_trained_model,
                dp_fl_trained_model,
                rel_epsilon=1e-4,
                abs_epsilon=1e-6,
            )

    def test_user_dp_variable_weights_same(self) -> None:
        """
        There are two cases when DP and NonDP should equal

        1) privacy is on, noise = 0, and clip norm is a large value
        2) privacy is on, noise * clip_norm << abs_epsilon

        Note:
            The current implementation has privacy on when noise >= 0 or clip_norm < inf
        """

        data_provider = get_fl_data_provider(
            num_examples=100,
            num_fl_users=10,
            examples_per_user=10,
            batch_size=5,
            model=DummyAlphabetFLModel(),
        )

        buffer_size = np.random.randint(1, 10)

        # sanity check DP == NonDP when privacy off
        error_msg = self._test_dp_no_dp_same_weighted_async(
            noise=-1,
            clip_norm=float("inf"),
            data_provider=data_provider,
            buffer_size=buffer_size,
        )
        assertEmpty(error_msg, msg=error_msg)
        # scenario 1
        error_msg = self._test_dp_no_dp_same_weighted_async(
            noise=0,
            clip_norm=1e6,
            data_provider=data_provider,
            buffer_size=buffer_size,
        )
        assertEmpty(error_msg, msg=error_msg)
        # scenario 2
        error_msg = self._test_dp_no_dp_same_weighted_async(
            noise=1e-14,
            clip_norm=1e6,
            data_provider=data_provider,
            buffer_size=buffer_size,
        )
        assertEmpty(error_msg, msg=error_msg)

    def test_user_dp_variable_weights_different(self) -> None:
        """
        Test when noise is some trivial number then
        DP model should not equal to NonDP model
        """
        buffer_size = np.random.randint(1, 10)
        data_provider = get_fl_data_provider(
            num_examples=10,
            num_fl_users=10,
            examples_per_user=1,
            batch_size=1,
            model=DummyAlphabetFLModel(),
        )
        is_different_msg = self._test_dp_no_dp_same_weighted_async(
            # noise between 0.1 and 1.0
            noise=max(0.1, np.random.sample()),
            clip_norm=10,
            data_provider=data_provider,
            buffer_size=buffer_size,
        )
        assertNotEmpty(is_different_msg, msg=is_different_msg)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

"""In this tutorial, we will train an image classifier with FLSim to simulate federated learning training environment.

With this tutorial, you will learn the following key components of FLSim:
1. Data loading
2. Model construction
3. Trainer construction

  Typical usage example:

  buck run @mode/dev-nosan papaya/toolkit/simulation/tutorials:cifar10_tutorial -- \\
  --config-file fblearner/flow/projects/papaya/examples/hydra_configs/cifar10_single_process.json
"""
import flsim.configs  # noqa
import hydra
import torch
from flsim.data.data_sharder import SequentialSharder
from flsim.examples.data.data_providers import FLVisionDataLoader, LEAFDataProvider
from flsim.examples.metrics_reporter.fl_metrics_reporter import MetricsReporter
from flsim.examples.models.cnn import SimpleConvNet
from flsim.examples.models.cv_model import FLModel
from flsim.interfaces.metrics_reporter import Channel
from flsim.utils.config_utils import maybe_parse_json_config
from hydra.utils import instantiate
from omegaconf import DictConfig, OmegaConf
from torchvision import transforms
from torchvision.datasets.cifar import CIFAR10


IMAGE_SIZE = 32


def build_data_provider(local_batch_size, examples_per_user, drop_last=False):

    transform = transforms.Compose(
        [
            transforms.Resize(IMAGE_SIZE),
            transforms.CenterCrop(IMAGE_SIZE),
            transforms.ToTensor(),
            transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
        ]
    )
    train_dataset = CIFAR10(
        root="../cifar10", train=True, download=True, transform=transform
    )
    test_dataset = CIFAR10(
        root="../cifar10", train=False, download=True, transform=transform
    )
    sharder = SequentialSharder(examples_per_shard=examples_per_user)
    fl_data_loader = FLVisionDataLoader(
        train_dataset, test_dataset, test_dataset, sharder, local_batch_size, drop_last
    )
    data_provider = LEAFDataProvider(fl_data_loader)
    print(f"Clients in total: {data_provider.num_users()}")
    return data_provider


def main(
    trainer_config,
    data_config,
    model_config,
    use_cuda_if_available=True,
):
    cuda_enabled = torch.cuda.is_available() and use_cuda_if_available
    device = torch.device(f"cuda:{0}" if cuda_enabled else "cpu")
    model = SimpleConvNet(in_channels=3, num_classes=10)
    global_model = FLModel(model, device)
    if cuda_enabled:
        global_model.fl_cuda()
    trainer = instantiate(trainer_config, model=global_model, cuda_enabled=cuda_enabled)
    print(f"Created {trainer_config._target_}")
    data_provider = build_data_provider(
        local_batch_size=data_config.local_batch_size,
        examples_per_user=data_config.examples_per_user,
        drop_last=False,
    )

    metrics_reporter = MetricsReporter([Channel.TENSORBOARD, Channel.STDOUT])

    final_model, eval_score = trainer.train(
        data_provider=data_provider,
        metric_reporter=metrics_reporter,
        num_total_users=data_provider.num_users(),
    )

    trainer.test(
        data_iter=data_provider.test_data(),
        metric_reporter=MetricsReporter([Channel.STDOUT]),
    )


@hydra.main(config_path=None, config_name="cifar10_tutorial")
def run(cfg: DictConfig):
    print(OmegaConf.to_yaml(cfg))

    trainer_config = cfg.trainer
    model_config = cfg.model
    data_config = cfg.data

    main(
        trainer_config,
        data_config,
        model_config,
    )


if __name__ == "__main__":
    cfg = maybe_parse_json_config()
    run(cfg)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from typing import Type

import numpy as np
from flsim.utils.timing.training_duration_distribution import (
    DurationDistributionConfig,
    PerExampleGaussianDurationDistributionConfig,
    PerUserGaussianDurationDistributionConfig,
)


class ConstantEventGenTestConfig:
    """This class specifies expected behavior for a training_simulator
    that uses an event generator where:
    a) Time between events is always 1 sec (or 1 unit of time)
    b) Training duration per example is 1 sec. 4 examples per user
       Thus, overall training duration is 4 secs
    c) There are 4 users, trained over 1 epoch
    This class is used both in a unit-test below, and in an integration
    test for ASyncTrainer

    In this case, training proceeds as follows:
    t=0, user1 STARTS training. model_seqnum=0
    t=1, user2 STARTS training. model_seqnum=0
    t=2, user3 STARTS training. model_seqnum=0
    t=3, user4 STARTS training. model_seqnum=0
    t=4, user1 FINISHES trng. Jobs pending=4. seqnum=1. SeqnumDiff=0
    t=5, user2 FINISHES trng. Jobs pending=3. seqnum=2. SeqnumDiff=1
    t=6, user3 FINISHES trng. Jobs pending=2. seqnum=3. SeqnumDiff=2
    t=7, user4 FINISHES trng. Jobs pending=1. seqnum=4. SeqnumDiff=3

    So, average number of pending jobs: mean([4, 3, 2, 1]) = 10/4 = 2.5
    SeqnumDiffs: [0,1,2,3]. Mean=6/4=1.5, SD = 1.18
    """

    num_examples_per_user = 4
    training_duration_distribution_config: Type[
        DurationDistributionConfig
    ] = PerExampleGaussianDurationDistributionConfig

    training_rate = 1
    training_duration_mean = 1
    training_duration_sd = 0
    num_users = 4

    pending_jobs = training_rate * num_users * training_rate
    seqnum_diffs = [0, 1, 2, 3]
    mean_pending_jobs = np.mean(pending_jobs)
    mean_seqnum_diff = np.mean(seqnum_diffs)
    sd_seqnum_diff = np.std(seqnum_diffs)


class ConstantEventGenTestConfigPerUserGaussian(ConstantEventGenTestConfig):
    """Same as ConstantEventGenTestConfig, but assumes that training duration
    distribution is a Per-User gaussian
    """

    # in parent class, training_duration_distrib was PerExampleGaussian
    # thus, parent training time per user = #examples-per-user * training_duration_mean
    # however, in this class, training_duration_distrb = PerUserGaussian
    # thus, training time per user = training_duration_mean
    # so we multiply training_duration_mean by #examples-per-user to keep
    # training duration constant
    num_examples_per_user = 1
    training_duration_mean = 4
    training_duration_distribution_config: Type[
        DurationDistributionConfig
    ] = PerUserGaussianDurationDistributionConfig


class PoissonEventGenTestConfig:
    """This class specifies expected behavior for a training_simulator
    that uses an event generator where:
    a) Time between events is 1 sec, Poisson distributed
    b) Training duration per example is 1/4 sec. 4 examples per user
       Thus, overall training duration is 1 sec
    c) There are 4 users,
    This class is used both in a unit-test below, and in an integration
    test for AsyncTrainer
    """

    num_examples_per_user = 4
    training_duration_distribution_config: Type[
        DurationDistributionConfig
    ] = PerExampleGaussianDurationDistributionConfig
    training_rate = 1
    training_duration_mean = 1 / 4
    training_duration_sd = 0
    num_users = 4

    mean_pending_jobs = 1.25
    mean_seqnum_diff = 0.25
    sd_seqnum_diff = 0.433


class PoissonEventGenTestConfigPerUserGaussian(PoissonEventGenTestConfig):
    """Same as PoissonEventGenTestConfig, but assumes that training duration
    distribution is a Per-User gaussian
    """

    num_examples_per_user = 1
    training_duration_distribution_config: Type[
        DurationDistributionConfig
    ] = PerUserGaussianDurationDistributionConfig
    training_duration_mean = 1



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from __future__ import annotations

import abc
import random
from collections import defaultdict
from dataclasses import dataclass
from typing import Any, Dict, Iterable, List, Tuple, TypeVar

import numpy as np
import torch
from flsim.utils.config_utils import fullclassname
from flsim.utils.config_utils import init_self_cfg
from omegaconf import MISSING
from torch.utils.data import Dataset


Shardable = TypeVar("Shardable", Iterable[Dict[str, Any]], Dataset)


class FLDataSharder(abc.ABC):
    """This class takes in a file, and partitions it into a list of datasets.
    It supports random partitioning, broadcasting,
    round-robining and sharding by a column
    """

    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=FLDataSharderConfig,
            **kwargs,
        )

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    @abc.abstractmethod
    def shard_for_row(self, csv_row: Dict[Any, Any]) -> List[Any]:
        """
        Determine which shard a row should belong to.
        """
        pass

    def shard_rows(self, data_rows: Shardable) -> Iterable[Tuple[str, Any]]:
        """Partition a set of rows into mulitple sets using a sharding strategy.

        Args:
            data_rows: Iterable[Dict[str, Any]]]: iterable over dictionary mapping column
            name to value.
        """
        shards = defaultdict(list)
        for one_row in data_rows:
            for shard_id in self.shard_for_row(one_row):
                shards[str(shard_id)].append(one_row)
        return shards.items()


class RandomSharder(FLDataSharder):
    """Splits training data randomly. num_shards should be specified.
    Assigns first num_shards rows to different shards to avoid empty shards."""

    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=RandomSharderConfig,
            **kwargs,
        )
        super().__init__(**kwargs)
        self._assignments: List[int] = list(range(self.cfg.num_shards))
        random.shuffle(self._assignments)

    def shard_for_row(self, csv_row: Dict[Any, Any]) -> List[int]:
        if self._assignments:
            return [self._assignments.pop(0)]
        # pyre-fixme[16]: `RandomSharder` has no attribute `cfg`.
        return [random.randint(0, self.cfg.num_shards - 1)]

    def shard_rows(self, data_rows: Shardable) -> Iterable[Tuple[str, Any]]:
        # sanity check to avoid empty shards
        shards = super().shard_rows(data_rows)
        assert (
            sum(1 for _shard in shards)
            # pyre-fixme[16]: `RandomSharder` has no attribute `cfg`.
            >= self.cfg.num_shards
        ), "number of rows must be at least the number of shards"
        return shards


class SequentialSharder(FLDataSharder):
    """Assign first N rows to shard A, the next N rows to shard B, and so
    on. Mostly used for testing purpose (e.g. sanity-check with
    conventional SGD training based on PyTorch Dataset and its batching
    mechanism).
    """

    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=SequentialSharderConfig,
            **kwargs,
        )
        super().__init__(**kwargs)
        self.filled_in_current_bucket = 0
        self.index = 0

    def shard_for_row(self, csv_row: Dict[Any, Any]) -> List[int]:
        # pyre-fixme[16]: `SequentialSharder` has no attribute `cfg`.
        if self.filled_in_current_bucket == self.cfg.examples_per_shard:
            self.index += 1
            self.filled_in_current_bucket = 1
        else:
            self.filled_in_current_bucket += 1
        return [self.index]


class BroadcastSharder(FLDataSharder):
    """Copy each training datum to all shards. num_shards should be specified"""

    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=BroadcastSharderConfig,
            **kwargs,
        )
        super().__init__(**kwargs)

    def shard_for_row(self, csv_row: Dict[Any, Any]) -> List[int]:
        # pyre-fixme[16]: `BroadcastSharder` has no attribute `cfg`.
        return list(range(self.cfg.num_shards))


class ColumnSharder(FLDataSharder):
    """Specify a column name used to shard.
    It should be the last column in the file,
    and sharding_column must be specified.
    """

    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=ColumnSharderConfig,
            **kwargs,
        )
        super().__init__(**kwargs)

    def shard_for_row(self, csv_row: Dict[Any, Any]) -> List[Any]:
        # pyre-fixme[16]: `ColumnSharder` has no attribute `cfg`.
        unwrapped_colindex = self.cfg.sharding_col
        if unwrapped_colindex.isdigit():
            unwrapped_colindex = int(unwrapped_colindex)
            assert unwrapped_colindex < len(csv_row), "Sharding index out of bounds: "
            f"{unwrapped_colindex}"

        shard_idx = csv_row[unwrapped_colindex]

        # shard_idx can be a 0-dim tensor when a Hive column is an integer.
        if isinstance(shard_idx, torch.Tensor):
            shard_idx = shard_idx.item()

        return [shard_idx]


class RoundRobinSharder(FLDataSharder):
    """Splits training in a round-robin fashion between all shards"""

    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=RoundRobinSharderConfig,
            **kwargs,
        )
        super().__init__(**kwargs)
        self._last_shard: int = -1

    def shard_for_row(self, csv_row: Dict[Any, Any]) -> List[int]:
        # pyre-fixme[16]: `RoundRobinSharder` has no attribute `cfg`.
        self._last_shard = (self._last_shard + 1) % self.cfg.num_shards
        return [self._last_shard]

    def shard_rows(self, data_rows: Shardable) -> Iterable[Tuple[str, Any]]:
        # sanity check to avoid empty shards
        shards = super().shard_rows(data_rows)
        assert (
            sum(1 for _shard in shards)
            # pyre-fixme[16]: `RoundRobinSharder` has no attribute `cfg`.
            >= self.cfg.num_shards
        ), "number of rows must be at least the number of shards"
        return shards


class PowerLawSharder(FLDataSharder):
    """
    Splits training data based on power law distribution with order
    alpha where alpha between [0.0, 1.0] to ensure right skewed

    This strategy will shard in a round robin at first to ensure all shard
    will get at least one example. After that, it will shard following the power
    law distribution
    """

    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=PowerLawSharderConfig,
            **kwargs,
        )
        super().__init__(**kwargs)
        assert 0.0 < self.cfg.alpha <= 1.0, "alpha must be in the interval (0, 1]"
        self._last_shard: int = -1
        self._weights = np.random.power(self.cfg.alpha, self.cfg.num_shards)
        # normalize to sum to 1.0
        self._weights = self._weights / sum(self._weights)
        self._choices = np.arange(0, self.cfg.num_shards)

    def shard_for_row(self, csv_row: Dict[Any, Any]) -> List[int]:
        # pyre-fixme[16]: `PowerLawSharder` has no attribute `cfg`.
        if self._last_shard < self.cfg.num_shards - 1:
            self._last_shard += 1
            return [self._last_shard]
        else:
            return [np.random.choice(self._choices, p=self._weights)]

    def shard_rows(self, data_rows: Shardable) -> Iterable[Tuple[str, Any]]:
        # sanity check to avoid empty shards
        shards = super().shard_rows(data_rows)
        assert (
            sum(1 for _shard in shards)
            # pyre-fixme[16]: `PowerLawSharder` has no attribute `cfg`.
            >= self.cfg.num_shards
        ), "number of rows must be at least the number of shards"
        return shards


@dataclass
class FLDataSharderConfig:
    _target_: str = MISSING
    _recursive_: bool = False


@dataclass
class RandomSharderConfig(FLDataSharderConfig):
    _target_: str = fullclassname(RandomSharder)
    num_shards: int = MISSING


@dataclass
class SequentialSharderConfig(FLDataSharderConfig):
    _target_: str = fullclassname(SequentialSharder)
    examples_per_shard: int = MISSING


@dataclass
class BroadcastSharderConfig(FLDataSharderConfig):
    _target_: str = fullclassname(BroadcastSharder)
    num_shards: int = MISSING


@dataclass
class ColumnSharderConfig(FLDataSharderConfig):
    _target_: str = fullclassname(ColumnSharder)
    sharding_col: str = MISSING


@dataclass
class RoundRobinSharderConfig(FLDataSharderConfig):
    _target_: str = fullclassname(RoundRobinSharder)
    num_shards: int = MISSING


@dataclass
class PowerLawSharderConfig(FLDataSharderConfig):
    _target_: str = fullclassname(PowerLawSharder)
    num_shards: int = MISSING
    alpha: float = MISSING



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

"""This file contains optimizers for the server.

The server expects an IServerOptimizer with two abstract methods: step and zero_grad.
This interface is similar to torch.optim.Optimizer.

  Typical usage example:

  optimizer = OptimizerType.create_optimizer(model=model, config=config)
  optimizer.step()
"""

from __future__ import annotations

import abc
from dataclasses import dataclass
from enum import Enum

import torch
import torch.nn as nn
from flsim.optimizers.layerwise_optimizers import LAMB, LARS
from flsim.utils.config_utils import fullclassname, init_self_cfg, is_target
from omegaconf import MISSING


class IServerOptimizer(abc.ABC):
    def __init__(self, *, model: nn.Module, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,  # pyre-fixme[10]: Name `__class__` is used but not defined.
            config_class=ServerOptimizerConfig,
            **kwargs,
        )
        self.model = model

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    @abc.abstractmethod
    @torch.no_grad()
    def step(self, closure):
        r"""Performs a single optimization step (parameter update).

        Args:
            closure (callable): A closure that reevaluates the model and
                returns the loss. Optional for most optimizers.

        .. note::
            Unless otherwise specified, this function should not modify the
            ``.grad`` field of the parameters.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def zero_grad(self, set_to_none: bool = False):
        r"""Sets the gradients of all optimized :class:`torch.Tensor` s to zero.

        Args:
            set_to_none (bool): instead of setting to zero, set the grads to None.
                This will in general have lower memory footprint, and can modestly improve performance.
                However, it changes certain behaviors. For example:
                1. When the user tries to access a gradient and perform manual ops on it,
                a None attribute or a Tensor full of 0s will behave differently.
                2. If the user requests ``zero_grad(set_to_none=True)`` followed by a backward pass, ``.grad``\ s
                are guaranteed to be None for params that did not receive a gradient.
                3. ``torch.optim`` optimizers have a different behavior if the gradient is 0 or None
                (in one case it does the step with a gradient of 0 and in the other it skips
                the step altogether).
        """
        raise NotImplementedError


# pyre-ignore[11] Annotation
class FedAvgWithLROptimizer(IServerOptimizer, torch.optim.SGD):
    def __init__(self, *, model: nn.Module, **kwargs) -> None:
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=FedAvgWithLROptimizerConfig,
            **kwargs,
        )

        IServerOptimizer.__init__(self, model=model, **kwargs)

        torch.optim.SGD.__init__(
            self,
            params=self.model.parameters(),
            # pyre-ignore[16] Undefined attribute
            lr=self.cfg.lr,
            momentum=self.cfg.momentum,
        )

    def step(self, closure=None):
        return torch.optim.SGD.step(self, closure)

    def zero_grad(self, set_to_none: bool = False):
        return torch.optim.SGD.zero_grad(self, set_to_none)


class FedAvgOptimizer(IServerOptimizer, torch.optim.SGD):
    def __init__(self, *, model: nn.Module, **kwargs) -> None:
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=FedAvgOptimizerConfig,
            **kwargs,
        )

        IServerOptimizer.__init__(self, model=model, **kwargs)

        torch.optim.SGD.__init__(
            self,
            params=self.model.parameters(),
            lr=1.0,
            momentum=0,
        )

    def step(self, closure=None):
        return torch.optim.SGD.step(self, closure)

    def zero_grad(self, set_to_none: bool = False):
        return torch.optim.SGD.zero_grad(self, set_to_none)


# pyre-ignore[11] Annotation
class FedAdamOptimizer(IServerOptimizer, torch.optim.Adam):
    def __init__(self, *, model: nn.Module, **kwargs) -> None:
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=FedAdamOptimizerConfig,
            **kwargs,
        )

        IServerOptimizer.__init__(self, model=model, **kwargs)

        torch.optim.Adam.__init__(
            self,
            params=self.model.parameters(),
            # pyre-ignore[16] Undefined attribute
            lr=self.cfg.lr,
            weight_decay=self.cfg.weight_decay,
            betas=(self.cfg.beta1, self.cfg.beta2),
            eps=self.cfg.eps,
        )

    def step(self, closure=None):
        return torch.optim.Adam.step(self, closure)

    def zero_grad(self, set_to_none: bool = False):
        return torch.optim.Adam.zero_grad(self, set_to_none)


class FedLARSOptimizer(IServerOptimizer, LARS):
    def __init__(self, *, model: nn.Module, **kwargs) -> None:
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=FedLARSOptimizerConfig,
            **kwargs,
        )

        IServerOptimizer.__init__(self, model=model, **kwargs)

        LARS.__init__(
            self,
            params=self.model.parameters(),
            # pyre-ignore[16] Undefined attribute
            lr=self.cfg.lr,
            weight_decay=self.cfg.weight_decay,
            beta=self.cfg.beta,
        )

    def step(self, closure=None):
        return LARS.step(self, closure)

    def zero_grad(self, set_to_none: bool = False):
        return LARS.zero_grad(self, set_to_none)


class FedLAMBOptimizer(IServerOptimizer, LAMB):
    def __init__(self, *, model: nn.Module, **kwargs) -> None:
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=FedLARSOptimizerConfig,
            **kwargs,
        )

        IServerOptimizer.__init__(self, model=model, **kwargs)

        LAMB.__init__(
            self,
            params=self.model.parameters(),
            # pyre-ignore[16] Undefined attribute
            lr=self.cfg.lr,
            weight_decay=self.cfg.weight_decay,
            beta1=self.cfg.beta1,
            beta2=self.cfg.beta2,
            eps=self.cfg.eps,
        )

    def step(self, closure=None):
        return LAMB.step(self, closure)

    def zero_grad(self, set_to_none: bool = False):
        return LAMB.zero_grad(self, set_to_none)


@dataclass
class ServerOptimizerConfig:
    _target_: str = MISSING
    _recursive_: bool = False


@dataclass
class FedAvgOptimizerConfig(ServerOptimizerConfig):
    _target_: str = fullclassname(FedAvgOptimizer)


@dataclass
class FedAvgWithLROptimizerConfig(ServerOptimizerConfig):
    _target_: str = fullclassname(FedAvgWithLROptimizer)
    lr: float = 0.001
    momentum: float = 0.0


@dataclass
class FedAdamOptimizerConfig(ServerOptimizerConfig):
    _target_: str = fullclassname(FedAdamOptimizer)
    lr: float = 0.001
    weight_decay: float = 0.00001
    beta1: float = 0.9
    beta2: float = 0.999
    eps: float = 1e-8


@dataclass
class FedLARSOptimizerConfig(ServerOptimizerConfig):
    _target_: str = fullclassname(FedLARSOptimizer)
    lr: float = 0.001
    weight_decay: float = 0.00001
    beta: float = 0.9


@dataclass
class FedLAMBOptimizerConfig(ServerOptimizerConfig):
    _target_: str = fullclassname(FedLAMBOptimizer)
    lr: float = 0.001
    weight_decay: float = 0.00001
    beta1: float = 0.9
    beta2: float = 0.999
    eps: float = 1e-8


class OptimizerType(Enum):
    fed_avg: str = FedAvgOptimizerConfig._target_
    fed_avg_with_lr: str = FedAvgWithLROptimizerConfig._target_
    fed_adam: str = FedAdamOptimizerConfig._target_
    fed_lamb: str = FedLAMBOptimizerConfig._target_
    fed_lars: str = FedLARSOptimizerConfig._target_

    @staticmethod
    def create_optimizer(
        model: nn.Module, config: ServerOptimizerConfig
    ) -> IServerOptimizer:
        if is_target(config, FedAvgWithLROptimizerConfig):
            return torch.optim.SGD(
                model.parameters(),
                # pyre-ignore[16] Undefined attribute
                lr=config.lr,
                # pyre-ignore[16] Undefined attribute
                momentum=config.momentum,
            )
        elif is_target(config, FedAdamOptimizerConfig):
            return torch.optim.Adam(
                model.parameters(),
                lr=config.lr,
                # pyre-ignore[16] Undefined attribute
                weight_decay=config.weight_decay,
                # pyre-ignore[16] Undefined attribute
                betas=(config.beta1, config.beta2),
                # pyre-ignore[16] Undefined attribute
                eps=config.eps,
            )
        elif is_target(config, FedLARSOptimizerConfig):
            # pyre-ignore[7]
            return LARS(
                model.parameters(),
                lr=config.lr,
                # pyre-ignore[16] Undefined attribute
                beta=config.beta,
                weight_decay=config.weight_decay,
            )

        elif is_target(config, FedLAMBOptimizerConfig):
            # pyre-ignore[7]
            return LAMB(
                model.parameters(),
                lr=config.lr,
                beta1=config.beta1,
                beta2=config.beta2,
                weight_decay=config.weight_decay,
                eps=config.eps,
            )
        elif is_target(config, FedAvgOptimizerConfig):
            return torch.optim.SGD(
                model.parameters(),
                lr=1.0,
                momentum=0,
            )
        else:
            raise ValueError(
                f"Optimizer type {config._target_} not found. Please update OptimizerType.create_optimizer"
            )



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from hydra.core.config_store import ConfigStore  #  @manual

from .secure_aggregator import FixedPointConfig

ConfigStore.instance().store(
    name="base_fixedpoint",
    node=FixedPointConfig,
    group="fixedpoint",
)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import copy
import math
import unittest
from dataclasses import dataclass
from tempfile import mkstemp
from typing import List, Type
from unittest.mock import MagicMock

import numpy as np
import torch
import torch.distributed as dist
import torch.multiprocessing as mp
from flsim.channels.message import Message
from flsim.channels.sketch_channel import SketchChannelConfig
from flsim.channels.sparse_mask_channel import SparseMaskChannelConfig
from flsim.clients.base_client import Client, ClientConfig
from flsim.privacy.common import PrivacySetting
from flsim.reducers.base_round_reducer import (
    ReductionType,
    RoundReducer,
    RoundReducerConfig,
)
from flsim.reducers.dp_round_reducer import DPRoundReducer, DPRoundReducerConfig
from flsim.reducers.secure_round_reducer import (
    SecureRoundReducer,
    SecureRoundReducerConfig,
)
from flsim.reducers.sketch_round_reducer import (
    SketchRoundReducer,
    SketchRoundReducerConfig,
)
from flsim.reducers.weighted_dp_round_reducer import (
    EstimatorType,
    WeightedDPRoundReducer,
    WeightedDPRoundReducerConfig,
)
from flsim.secure_aggregation.secure_aggregator import FixedPointConfig
from flsim.tests import utils
from flsim.utils.async_trainer.async_example_weights import EqualExampleWeightConfig
from flsim.utils.async_trainer.async_staleness_weights import (
    PolynomialStalenessWeightConfig,
)
from flsim.utils.async_trainer.async_weights import AsyncWeight, AsyncWeightConfig
from flsim.utils.count_sketch import CountSketch
from flsim.utils.distributed.fl_distributed import FLDistributedUtils
from hydra.errors import HydraException
from hydra.utils import instantiate
from libfb.py import testutil
from omegaconf import OmegaConf


def init_process(rank, world_size, reducer, models, file_loc, pipe, backend="gloo"):
    """
    Initialize the distributed environment for multi-process testing, and
    runs a simple reduction task and returns the mean of all the parameters
    int reduced module.
    """
    dist.init_process_group(
        backend, init_method=f"file://{file_loc}", rank=rank, world_size=world_size
    )
    reducer.reset(models[0])
    for i, m in enumerate(models):
        if i % world_size == rank:
            reducer.collect_update(m, float(i + 1))
    reducer.reduce()
    sums, weights = 0.0, 0.0
    all_sum = [(p.sum(), p.numel()) for p in reducer.reduced_module.parameters()]
    for s, w in all_sum:
        sums += float(s)
        weights += float(w)
    pipe.send(sums / weights)
    dist.destroy_process_group()


def run_reduction_test(reducer, num_processes=1, num_models=4):
    """
    Used in multiprocess test only.
    Runs a simple scenario in multiple processes.
    Models are sequentially initilized with
    a constant value for params, i.e. 1, 2, ..., num_models.
    """
    _, tmpfile = mkstemp(dir="/tmp")
    pipe_out, pipe_in = mp.Pipe(False)
    # reducer.reduced_module.share_memory()
    models = [utils.SampleNet(utils.TwoFC()) for _ in range(num_models)]
    for i, m in enumerate(models):
        m.fl_get_module().fill_all(float(i + 1))
    processes = []
    results = []
    FLDistributedUtils.WORLD_SIZE = num_processes
    for pid in range(num_processes):
        p = mp.Process(
            target=init_process,
            args=(pid, num_processes, reducer, models, tmpfile, pipe_in),
        )
        p.start()
        processes.append(p)
        results.append(pipe_out)
    for p in processes:
        p.join()
    res = [r.recv() for r in results]
    return res


class RoundReducerTestBase(unittest.TestCase):
    def _fake_client(self, global_value, client_value, weight):
        clnt = Client(dataset=None, **OmegaConf.structured(ClientConfig()))

        def fill(model, *args):
            model.fl_get_module().fill_all(global_value - client_value)
            return model, weight

        clnt.generate_local_update = MagicMock(side_effect=fill)
        return clnt

    def _create_fake_clients(
        self, global_param_value, num_clients, client_param_value, client_weight
    ) -> List[Client]:
        # initialize clients, each with model parameters equal to client_param_values
        return [
            self._fake_client(
                global_value=global_param_value,
                client_value=client_param_value,
                weight=client_weight,
            )
            for _ in range(num_clients)
        ]


class RoundReducerTest(RoundReducerTestBase):
    def get_round_reducer(
        self,
        model=None,
        reduction_type=ReductionType.WEIGHTED_AVERAGE,
        reset=True,
    ):
        model = model or utils.SampleNet(utils.TwoFC())
        round_reducer = RoundReducer(
            **OmegaConf.structured(RoundReducerConfig(reduction_type=reduction_type)),
            global_model=model,
        )
        return round_reducer

    def test_reset(self):
        rr = self.get_round_reducer()
        mismatched = utils.model_parameters_equal_to_value(rr.reduced_module, 0.0)
        self.assertEqual(mismatched, "", mismatched)
        # do it again
        rr.reduced_module.fill_all(1.0)
        rr.reset(utils.SampleNet(utils.TwoFC()))
        mismatched = utils.model_parameters_equal_to_value(rr.reduced_module, 0.0)
        self.assertEqual(mismatched, "", mismatched)

    def test_receive_through_channel(self):
        # expected channel effects,
        rr = self.get_round_reducer()
        model = utils.SampleNet(utils.TwoFC())
        # check channel is pass through
        # TODO modify when there is actuall channel
        model2 = rr.receive_through_channel(model)
        mismatched = utils.verify_models_equivalent_after_training(model2, model)
        self.assertEqual(mismatched, "", mismatched)

    def test_update_reduced_module(self):
        model = utils.SampleNet(utils.TwoFC())
        rr = self.get_round_reducer(model)
        model.fl_get_module().fill_all(0.2)
        rr.update_reduced_module(model.fl_get_module(), 3.0)
        model.fl_get_module().fill_all(0.3)
        rr.update_reduced_module(model.fl_get_module(), 2.0)
        mismatched = utils.model_parameters_equal_to_value(
            rr.reduced_module, 3 * 0.2 + 2 * 0.3
        )
        self.assertEqual(mismatched, "", mismatched)

    def test_collect_update(self):
        param_values = [0.1 * i for i in range(100)]
        weights = [i % 10 for i in range(100)]
        global_param = 1.0
        clients = [
            self._fake_client(global_param, p, w) for p, w in zip(param_values, weights)
        ]
        ref_model = utils.SampleNet(utils.TwoFC())
        rr = self.get_round_reducer(ref_model)
        for clnt in clients:
            model, weight = clnt.generate_local_update(utils.SampleNet(utils.TwoFC()))
            rr.collect_update(model, weight)
        expected_sum_weights = sum(weights)
        expected_param_values = sum(
            (global_param - param_value) * w
            for param_value, w in zip(param_values, weights)
        )
        experiment_model, experiment_weight = rr.current_results
        self.assertAlmostEqual(expected_sum_weights, experiment_weight, 5)
        mismatched = utils.model_parameters_equal_to_value(
            experiment_model, expected_param_values
        )
        self.assertEqual(mismatched, "", mismatched)

    def test_reduction_types_sum(self):
        model = utils.SampleNet(utils.TwoFC())
        rr = self.get_round_reducer(model, reduction_type=ReductionType.SUM)
        results = run_reduction_test(rr, num_processes=1, num_models=2)
        value_expected = float(sum(i + 1 for i in range(2)))
        for r in results:
            self.assertAlmostEqual(r, value_expected, 5)

    def test_reduction_types_avg(self):
        model = utils.SampleNet(utils.TwoFC())
        rr = self.get_round_reducer(model, reduction_type=ReductionType.AVERAGE)
        results = run_reduction_test(rr, num_processes=2, num_models=4)
        value_expected = sum(i + 1 for i in range(4)) / 4
        for r in results:
            self.assertAlmostEqual(r, value_expected, 5)

    def test_reduction_types_weighted_sum(self):
        model = utils.SampleNet(utils.TwoFC())
        rr = self.get_round_reducer(model, reduction_type=ReductionType.WEIGHTED_SUM)
        results = run_reduction_test(rr, num_processes=3, num_models=6)
        value_expected = float(sum((i + 1) ** 2 for i in range(6)))
        for r in results:
            self.assertAlmostEqual(r, value_expected, 5)

    def test_reduction_types_weighted_avg(self):
        model = utils.SampleNet(utils.TwoFC())
        rr = self.get_round_reducer(
            model, reduction_type=ReductionType.WEIGHTED_AVERAGE
        )
        results = run_reduction_test(rr, num_processes=4, num_models=8)
        value_expected = float(sum((i + 1) ** 2 for i in range(8))) / sum(
            i + 1 for i in range(8)
        )
        for r in results:
            self.assertAlmostEqual(r, value_expected, 5)

    def test_multiprocess_reduce(self):
        # test multi processing.
        model = utils.SampleNet(utils.TwoFC())
        num_models = 4
        value_expected = float(sum((i + 1) ** 2 for i in range(num_models))) / sum(
            i + 1 for i in range(num_models)
        )
        # test 1 process
        r1 = self.get_round_reducer(model, reset=False)
        results = run_reduction_test(r1, num_processes=1, num_models=num_models)
        for r in results:
            self.assertAlmostEqual(r, value_expected, 5)
        # test 4 processes
        r2 = self.get_round_reducer(model)
        results = run_reduction_test(r2, num_processes=2, num_models=num_models)
        for r in results:
            self.assertAlmostEqual(r, value_expected, 5)

    def test_logging_level(self):
        rr = self.get_round_reducer()
        self.assertTrue(utils.check_inherit_logging_level(rr, 50))
        self.assertTrue(utils.check_inherit_logging_level(rr, 10))


class SecureRoundReducerTest(RoundReducerTestBase):
    def get_round_reducer(
        self,
        model,
        fixedpoint,
    ):
        model = model or utils.SampleNet(utils.TwoFC())
        round_reducer = SecureRoundReducer(
            **OmegaConf.structured(
                SecureRoundReducerConfig(
                    reduction_type=ReductionType.WEIGHTED_AVERAGE,
                    fixedpoint=fixedpoint,
                )
            ),
            global_model=model,
        )
        return round_reducer

    def test_secure_aggregator_reducer_init(self):
        """
        Tests whether secure aggeragtion object is initiated
        correctly in round reducer
        """
        model = utils.SampleNet(utils.TwoFC())
        # test secure aggregation with flat FP config
        fixedpoint = FixedPointConfig(num_bytes=2, scaling_factor=100)
        r1 = self.get_round_reducer(model, fixedpoint=fixedpoint)
        self.assertEqual(len(r1.secure_aggregator.converters), 4)
        self.assertEqual(  # verify an arbitrary layer of the model
            r1.secure_aggregator.converters["fc2.bias"].scaling_factor, 100
        )

    def test_secure_aggregator_reducer_update_reduced_module(self):
        """
        Tests whether secure aggeragtion operations work correctly
        when a model update is received and reduced module is updated
        """
        model = utils.SampleNet(utils.TwoFC())

        fixedpoint = FixedPointConfig(num_bytes=2, scaling_factor=100)
        # hence minValue = -32768, maxValue = 32767
        rr = self.get_round_reducer(model, fixedpoint=fixedpoint)

        model.fl_get_module().fill_all(7.2345)
        rr.update_reduced_module(model.fl_get_module(), 3.0)
        model.fl_get_module().fill_all(-3.45612)
        rr.update_reduced_module(model.fl_get_module(), 7.0)
        mismatched = utils.model_parameters_equal_to_value(
            rr.reduced_module, -249.0
        )  # round(7.2345*100*3 - 3.45612*100*7) = round(-248.934)
        self.assertEqual(mismatched, "", mismatched)

        rr.reduce()
        mismatched = utils.model_parameters_equal_to_value(
            rr.reduced_module, -0.2490  # -249.0 / 100 / (3.0 + 0.7)
        )
        self.assertEqual(mismatched, "", mismatched)

    def test_secure_aggregator_reducer_reduce_large_range(self):
        """
        Tests whether secure aggeragtion operations work correctly
        when the reduce() method is called, and when the num_bytes is
        big, so we do not have a possible fixedpoint overflow
        """
        # TODO: revisit when we implement overflow in aggregation T96117010
        ref_model = utils.SampleNet(utils.TwoFC())
        fixedpoint = FixedPointConfig(num_bytes=4, scaling_factor=10)
        rr = self.get_round_reducer(ref_model, fixedpoint=fixedpoint)

        global_param = 8.0
        client_param = 2.123
        num_clients = 50

        clients = self._create_fake_clients(
            global_param_value=global_param,
            num_clients=num_clients,
            client_param_value=client_param,
            client_weight=1.0,
        )
        for client in clients:
            delta, weight = client.generate_local_update(utils.SampleNet(utils.TwoFC()))
            rr.collect_update(delta, weight)
        rr.reduce()
        # update is 5.877. In current fixedpoint setup, it is 59.
        # when we reduce, we convert from fixedpoint to floating point, so 59 -> 5.9
        expected_param_values = (5.9 * num_clients) / num_clients

        model_after_noise, sum_weights = rr.current_results
        mismatched = utils.model_parameters_equal_to_value(
            model_after_noise, expected_param_values
        )
        self.assertEqual(mismatched, "", mismatched)

    def test_secure_aggregator_reducer_reduce_small_range(self):
        """
        Tests whether secure aggeragtion operations work correctly
        when the reduce() method is called, and when the num_bytes is
        small so we have possible fixedpoint overflows
        """
        # TODO: revisit when we implement overflow in aggregation T96117010
        ref_model = utils.SampleNet(utils.TwoFC())
        fixedpoint = FixedPointConfig(num_bytes=1, scaling_factor=100)
        # the range of numbers would be (-128,127)
        rr = self.get_round_reducer(ref_model, fixedpoint=fixedpoint)

        global_param = 8.0
        client_param = 2.123
        num_clients = 50

        clients = self._create_fake_clients(
            global_param_value=global_param,
            num_clients=num_clients,
            client_param_value=client_param,
            client_weight=1.0,
        )
        for client in clients:
            delta, weight = client.generate_local_update(utils.SampleNet(utils.TwoFC()))
            rr.collect_update(delta, weight)
        rr.reduce()
        # update is 5.877. In current fixedpoint setup, it is 127, as 587.7 > 127
        # when we reduce, we convert from fixedpoint to floating point, so 127 -> 1.27
        expected_param_values = (1.27 * num_clients) / num_clients

        model_after_noise, sum_weights = rr.current_results

        mismatched = utils.model_parameters_equal_to_value(
            model_after_noise, expected_param_values
        )
        self.assertEqual(mismatched, "", mismatched)


class DPRoundReducerTest(RoundReducerTestBase):
    def _get_dp_round_reducer(
        self,
        ref_model=None,
        clipping_value=99999.99,
        reduction_type=ReductionType.AVERAGE,
        noise_multiplier=0,
        num_users_per_round: int = 1,
        total_number_of_users: int = 1,
        reset=True,
    ):
        ref_model = ref_model or utils.SampleNet(utils.TwoFC())
        privacy_setting = PrivacySetting(
            noise_multiplier=noise_multiplier, clipping_value=clipping_value
        )
        dp_rr = DPRoundReducer(
            **OmegaConf.structured(
                DPRoundReducerConfig(
                    reduction_type=reduction_type, privacy_setting=privacy_setting
                )
            ),
            global_model=ref_model,
            num_users_per_round=num_users_per_round,
            total_number_of_users=total_number_of_users,
        )
        if reset:
            dp_rr.reset(ref_model)
        return dp_rr

    def _create_ref_model(self, ref_model_param_value):
        ref_model = utils.SampleNet(utils.TwoFC())
        ref_model.fl_get_module().fill_all(ref_model_param_value)
        return ref_model

    def test_dp_off(self):
        ref_model = self._create_ref_model(ref_model_param_value=3.0)
        # when clipping_value is inf, sensetivity is inf -> dp not supported
        dp_rr = self._get_dp_round_reducer(
            ref_model, clipping_value=float("inf"), noise_multiplier=0
        )
        self.assertFalse(dp_rr.privacy_on)
        # noise < 0 means no dp
        dp_rr = self._get_dp_round_reducer(
            ref_model, clipping_value=10.0, noise_multiplier=-1
        )
        self.assertFalse(dp_rr.privacy_on)

    def test_collect_update_with_clipping(self):
        """
        Tests whether the model updates associated with the new models sent
        from clients are clipped correctly.
        """
        num_clients = 100
        global_value = 5.0
        clients = self._create_fake_clients(
            global_param_value=global_value,
            num_clients=num_clients,
            client_param_value=3.0,
            client_weight=1.0,
        )
        ref_model = self._create_ref_model(ref_model_param_value=global_value)

        dp_rr = self._get_dp_round_reducer(
            ref_model,
            clipping_value=6.0,
            num_users_per_round=num_clients,
            total_number_of_users=num_clients,
        )
        for client in clients:
            delta, weight = client.generate_local_update(utils.SampleNet(utils.TwoFC()))
            dp_rr.collect_update(delta, weight)
            """
            delta = global (all 5.0) - local (all 3.0) = all 2.0
            delta norm = sqrt(num_params*delta^2)=sqrt(21*2^2)=sqrt(84)= 9.16515138991168
            and this will be clipped to clipping_value of 6.0, which
            means that the parameters of the clipped update will be all equal
            to sqrt(36/21)= 1.309307341415954
            """

        expected_param_values = (1.309307341415954) * num_clients
        collected_model_updates, _ = dp_rr.current_results
        mismatched = utils.model_parameters_equal_to_value(
            collected_model_updates, expected_param_values
        )
        self.assertEqual(mismatched, "", mismatched)

    def test_clipping_when_noise_zero(self):
        """
        Tests when noise multiplier is zero, calling add_noise() in reduce()
        does not change the model after clipping.
        """
        num_clients = 50
        global_value = 8.0
        clients = self._create_fake_clients(
            global_param_value=global_value,
            num_clients=num_clients,
            client_param_value=2.0,
            client_weight=1.0,
        )
        ref_model = self._create_ref_model(ref_model_param_value=global_value)

        dp_rr = self._get_dp_round_reducer(
            ref_model,
            clipping_value=15.0,
            noise_multiplier=0,
            num_users_per_round=num_clients,
            total_number_of_users=num_clients,
        )
        for client in clients:
            delta, weight = client.generate_local_update(utils.SampleNet(utils.TwoFC()))

            dp_rr.collect_update(delta, weight)
            """
            update = global (all 8.0) - local (all 2.0) = all 6.0
            update norm = sqrt(num_params*delta^2)=sqrt(21*6^2)=sqrt(756)= 27.49545416973504
            and this will be clipped to clipping_value of 15, which
            means that the parameters of the clipped update will be all equal
            to sqrt(15^2/21)= 3.273268353539886
            """

        dp_rr.reduce()
        # asserts callnig add_noise did not change anything
        expected_param_values = ((3.273268353539886 * num_clients) / num_clients) + 0

        model_after_noise, sum_weights = dp_rr.current_results
        mismatched = utils.model_parameters_equal_to_value(
            model_after_noise, expected_param_values
        )
        self.assertEqual(mismatched, "", mismatched)

    def test_noise_when_clipping_large_value(self):
        """
        Tests 2 things: 1) whether clipping does not happen when
        clipping threshold is set to a large value, 2) whether we get
        a different model when we add noise and clipping is ineffective
        (clipping threshold is set to a large value).
        """
        num_clients = 20
        global_value = 5.0
        clients = self._create_fake_clients(
            global_param_value=global_value,
            num_clients=num_clients,
            client_param_value=3.0,
            client_weight=1.0,
        )
        ref_model = self._create_ref_model(ref_model_param_value=global_value)

        ref_model_before = copy.deepcopy(ref_model)
        dp_rr = self._get_dp_round_reducer(
            ref_model,
            clipping_value=10.0,
            num_users_per_round=num_clients,
            total_number_of_users=num_clients,
        )
        for client in clients:
            delta, weight = client.generate_local_update(utils.SampleNet(utils.TwoFC()))
            dp_rr.collect_update(delta, weight)
            """
            update = global (all 5.0) - local (all 3.0) = all 2.0
            update norm = sqrt(num_params*delta^2)=sqrt(21*2^2)=sqrt(84)= 9.16515138991168
            and this will not be clipped, because the clipping_value
            is set as a larger value (10 > 9.16515138991168). So the parameters
            of the model update will not change and all be equal to 2.
            """

        # asserts clipping does not happen
        expected_param_values = 2.0 * num_clients
        collected_model_updates, _ = dp_rr.current_results
        mismatched = utils.model_parameters_equal_to_value(
            collected_model_updates, expected_param_values
        )
        self.assertEqual(mismatched, "", mismatched)

        dp_rr.reduce()
        ref_module_after_noise, _ = dp_rr.current_results
        # asserts by adding noise, we get a different model
        mismatched = utils.verify_models_equivalent_after_training(
            ref_model_before.fl_get_module(), ref_module_after_noise
        )
        self.assertNotEqual(mismatched, "")

    def test_noise_added_correctly(self):
        """
        Tests that the noise is added correctly to model.
        """
        num_clients = 100
        global_value = 5.0
        clients = self._create_fake_clients(
            global_param_value=global_value,
            num_clients=num_clients,
            client_param_value=3.0,
            client_weight=1.0,
        )
        ref_model = self._create_ref_model(ref_model_param_value=global_value)

        dp_rr = self._get_dp_round_reducer(
            ref_model,
            clipping_value=7.0,
            num_users_per_round=num_clients,
            total_number_of_users=num_clients,
        )
        for client in clients:
            delta, weight = client.generate_local_update(utils.SampleNet(utils.TwoFC()))
            client.compute_delta(ref_model, delta, delta)
            dp_rr.collect_update(delta, weight)
            """
            update = global (all 5.0) - local (all 3.0) = all 2.0
            update norm = sqrt(num_params*delta^2)=sqrt(21*2^2)=sqrt(84)= 9.16515138991168
            and this will be clipped to clipping_value of 7, which
            means that the parameters of the clipped update will be all equal
            to sqrt(49/21)= 1.527525231651947
            """

        dp_rr.privacy_engine._generate_noise = MagicMock(return_value=0.8)
        expected_param_values = ((1.527525231651947 * num_clients) / num_clients) + 0.8
        dp_rr.reduce()
        ref_module_after_noise, _ = dp_rr.current_results
        mismatched = utils.model_parameters_equal_to_value(
            ref_module_after_noise, expected_param_values
        )
        self.assertEqual(mismatched, "", mismatched)

    def test_multiprocess_dp_all_processes_the_same(self):
        # test multi processing.
        model = utils.SampleNet(utils.TwoFC())
        num_models = 4
        # test 4 processes
        r = self._get_dp_round_reducer(
            model,
            clipping_value=1.0,
            reduction_type=ReductionType.AVERAGE,
            noise_multiplier=1,
            num_users_per_round=4,
            total_number_of_users=4,
            reset=False,
        )
        results = run_reduction_test(r, num_processes=4, num_models=num_models)
        same_value = results[0]
        for r in results:
            self.assertAlmostEqual(r, same_value, 5)


@dataclass
class WeightReducerTestSetting:
    num_clients: int = 10
    clients_per_round: int = 5
    clipping_value: float = 1.0
    max_staleness: int = 100
    noise: float = 1.0
    max_weight: float = 1.0
    min_weight: float = 0.0
    mean_weight: float = 0.5


class WeightedDPRoundReducerTest(DPRoundReducerTest):
    def _get_reducer(
        self,
        ref_model=None,
        clipping_value=1e10,
        reduction_type=ReductionType.WEIGHTED_SUM,
        estimator_type=EstimatorType.UNBIASED,
        noise_multiplier=0,
        num_users_per_round: int = 1,
        total_number_of_users: int = 1,
        max_weight: float = 10,
        min_weight: float = 1e-6,
        mean_weight: float = 1e-6,
    ):
        ref_model = ref_model or utils.SampleNet(utils.TwoFC())
        privacy_setting = PrivacySetting(
            noise_multiplier=noise_multiplier, clipping_value=clipping_value
        )
        reducer = WeightedDPRoundReducer(
            **OmegaConf.structured(
                WeightedDPRoundReducerConfig(
                    reduction_type=reduction_type,
                    privacy_setting=privacy_setting,
                    estimator_type=estimator_type,
                    max_weight=max_weight,
                    min_weight=min_weight,
                    mean_weight=mean_weight,
                )
            ),
            global_model=ref_model,
            num_users_per_round=num_users_per_round,
            total_number_of_users=total_number_of_users,
        )
        return reducer

    def _get_async_weight(self, exponent, avg_staleness=0):
        return AsyncWeight(
            **OmegaConf.structured(
                AsyncWeightConfig(
                    staleness_weight=PolynomialStalenessWeightConfig(
                        exponent=exponent, avg_staleness=avg_staleness
                    ),
                    example_weight=EqualExampleWeightConfig(),
                )
            )
        )

    def _get_num_params(self, model):
        return sum(p.numel() for p in model.parameters())

    def _reduce_weighted_models(
        self,
        global_model,
        settings: WeightReducerTestSetting,
        reduction_type: ReductionType,
        estimator_type: EstimatorType = EstimatorType.UNBIASED,
        global_param: float = 1.0,
        client_param: float = 1.0,
        client_weight: float = 1.0,
    ):
        clients = self._create_fake_clients(
            global_param_value=global_param,
            num_clients=settings.num_clients,
            client_param_value=client_param,
            client_weight=client_weight,
        )

        reducer = self._get_reducer(
            global_model,
            reduction_type=reduction_type,
            clipping_value=settings.clipping_value,
            estimator_type=estimator_type,
            num_users_per_round=settings.clients_per_round,
            total_number_of_users=settings.num_clients,
            max_weight=settings.max_weight,
            min_weight=settings.min_weight,
            mean_weight=settings.mean_weight,
        )
        async_weight = self._get_async_weight(exponent=0.5)

        weights = []
        for client in clients:
            delta, model_weight = client.generate_local_update(
                utils.SampleNet(utils.TwoFC())
            )
            staleness = np.random.randint(1, settings.max_staleness)
            weight = async_weight.weight(num_examples=model_weight, staleness=staleness)
            self.assertTrue(0.0 <= weight <= 1.0)

            reducer.collect_update(delta, weight)
            weights.append(weight)

        reducer.privacy_engine._generate_noise = MagicMock(return_value=settings.noise)
        reducer.reduce()
        return reducer, weights

    def _test_weighted_avg_reduction(
        self, estimator_type, global_param, client_param, max_clip_norm
    ):
        delta = global_param - client_param
        global_model = self._create_ref_model(ref_model_param_value=global_param)
        num_params = self._get_num_params(global_model.fl_get_module())
        user_norm = math.sqrt(num_params * delta ** 2)

        settings = WeightReducerTestSetting(
            noise=np.random.sample(),
            clipping_value=max_clip_norm,
            max_weight=1,
            min_weight=1e-6,
            mean_weight=1e-6,
        )
        reducer, model_updates = self._reduce_weighted_models(
            global_model=global_model,
            settings=settings,
            reduction_type=ReductionType.WEIGHTED_AVERAGE,
            estimator_type=estimator_type,
            client_param=client_param,
            client_weight=10,
            global_param=global_param,
        )
        if max_clip_norm <= user_norm:
            expected_param_values = (
                delta * (settings.clipping_value / user_norm) + settings.noise
            )
        else:
            expected_param_values = delta + settings.noise

        ref_module_after_noise, _ = reducer.current_results
        return utils.model_parameters_equal_to_value(
            ref_module_after_noise, expected_param_values
        )

    def test_clipped_models_weighted_sum(self):
        """
        Test when models get clipped with weighted sum

        1. Compute the expected per user L2 Norm
        2. Set the clipping threshold to be between 1 and user norm
        3. Expected model param should be
        global = init_global - sum(clipped norms * weights) - noise
        """

        global_param = 5
        client_param = 1
        delta = global_param - client_param

        global_model = self._create_ref_model(ref_model_param_value=global_param)
        num_params = self._get_num_params(global_model.fl_get_module())
        user_norm = math.sqrt(num_params * delta ** 2)
        settings = WeightReducerTestSetting(
            num_clients=10,
            clients_per_round=10,
            noise=np.random.sample(),
            # clipping value is between 1 and user norm
            clipping_value=np.random.randint(1, user_norm),
        )
        reducer, weights = self._reduce_weighted_models(
            global_model=global_model,
            settings=settings,
            reduction_type=ReductionType.WEIGHTED_SUM,
            client_param=client_param,
            global_param=global_param,
        )

        clipped_deltas = math.sqrt(settings.clipping_value ** 2 / num_params)

        model_updates = sum((w * clipped_deltas for w in weights))
        expected_param_values = model_updates + settings.noise

        ref_module_after_noise, _ = reducer.current_results
        mismatched = utils.model_parameters_equal_to_value(
            ref_module_after_noise, expected_param_values
        )
        self.assertEqual(mismatched, "", mismatched)

    def test_clipped_models_weighted_avg_with_biased_estimator(self):
        """
        Test when models get clipped with weighted avg with biased estimator
        where the sensitivity = (clipping_value * max_weight) / (mean_weight * num_clients_per_round)

        1. Compute the expected per user L2 Norm
        2. Set the clipping threshold to be between 1 and user norm
        3. Expected model param should be
            delta = init_global - client
            clipped_norm = min(max_clip_norm / user_norm, 1.0)
            global = delta * clipped_norms - noise
        """
        global_param = 5
        client_param = 1
        mismatched = self._test_weighted_avg_reduction(
            EstimatorType.BIASED, global_param, client_param, max_clip_norm=1
        )
        self.assertEqual(mismatched, "", mismatched)

    def test_clipped_models_weighted_avg_with_unbiased_estimator(self):
        """
        Test when models get clipped with weighted avg with unbiased estimator
        where the sensitivity = (clipping_value * max_weight) / (min_weight * users_per_round)

        1. Compute the expected per user L2 Norm
        2. Set the clipping threshold to be between 1 and user norm
        3. Expected model param should be
            delta = init_global - client
            clipped_norm = min(max_clip_norm / user_norm, 1.0)
            global = delta * clipped_norms - noise
        4.
        """
        global_param = 5
        client_param = 1
        mismatched = self._test_weighted_avg_reduction(
            EstimatorType.UNBIASED, global_param, client_param, max_clip_norm=1
        )
        self.assertEqual(mismatched, "", mismatched)

    def test_unclipped_models_weighted_avg_with_biased_estimator(self):
        """
        Test when max_clip_norm is greater than user norm with weighted avg

        When the models are unclipped then the expected global model is
            delta = init_global - client
            global = init_global - delta - noise
        """
        global_param = 5
        client_param = 1

        mismatched = self._test_weighted_avg_reduction(
            EstimatorType.BIASED, global_param, client_param, max_clip_norm=100
        )
        self.assertEqual(mismatched, "", mismatched)

    def test_unclipped_models_weighted_avg_with_unbiased_estimator(self):
        """
        Test when max_clip_norm is greater than user norm with weighted avg

        When the models are unclipped then the expected global model is
            delta = init_global - client
            global = init_global - delta - noise
        """
        global_param = 5
        client_param = 1

        mismatched = self._test_weighted_avg_reduction(
            EstimatorType.UNBIASED, global_param, client_param, max_clip_norm=100
        )
        self.assertEqual(mismatched, "", mismatched)

    def test_unclipped_models_weighted_sum(self):
        """
        Test when max_clip_norm is greater than user norm with weighted sum

        When the models are unclipped then the expected global model is
            delta = init_global - client
            global = init_global - sum(delta * weights) - noise
        """
        global_param = np.random.randint(2, 10)
        client_param = np.random.randint(1, global_param)
        delta = global_param - client_param

        global_model = self._create_ref_model(ref_model_param_value=global_param)
        num_params = self._get_num_params(global_model.fl_get_module())
        user_norm = math.sqrt(num_params * delta ** 2)
        settings = WeightReducerTestSetting(
            num_clients=10,
            clients_per_round=10,
            noise=np.random.sample(),
            # clipping value is greater than user norm
            clipping_value=user_norm + 1,
        )
        reducer, weights = self._reduce_weighted_models(
            global_model=global_model,
            settings=settings,
            reduction_type=ReductionType.WEIGHTED_SUM,
            client_param=client_param,
            global_param=global_param,
        )

        model_updates = sum((w * delta for w in weights))
        expected_param_values = model_updates + settings.noise
        ref_module_after_noise, _ = reducer.current_results
        mismatched = utils.model_parameters_equal_to_value(
            ref_module_after_noise, expected_param_values
        )
        self.assertEqual(mismatched, "", mismatched)

    def test_weighted_dp_multiprocess_same(self):
        """
        Multiprocess test for weighted DP reducer
        """
        model = utils.SampleNet(utils.TwoFC())

        # test 4 processes
        r4 = self._get_dp_round_reducer(
            model,
            clipping_value=1.0,
            reduction_type=ReductionType.WEIGHTED_AVERAGE,
            noise_multiplier=1,
            num_users_per_round=4,
            total_number_of_users=4,
            reset=False,
        )
        results_4 = run_reduction_test(r4, num_processes=4, num_models=4)
        same_value = results_4[0]
        for r in results_4:
            self.assertAlmostEqual(r, same_value, places=5)


class FLRoundReducerTest(testutil.BaseFacebookTestCase):
    def setUp(self) -> None:
        super().setUp()

    @testutil.data_provider(
        lambda: (
            {
                "config": RoundReducerConfig(),
                "expected_type": RoundReducer,
            },
        )
    )
    def test_reducer_creation_from_config(self, config: Type, expected_type: Type):
        ref_model = utils.SampleNet(utils.TwoFC())
        reducer = instantiate(config, global_model=ref_model)
        self.assertIsInstance(reducer, expected_type)


class SketchRoundReducerTest(RoundReducerTestBase):
    def get_round_reducer(
        self,
        model=None,
        reduction_type=ReductionType.WEIGHTED_AVERAGE,
        reset=True,
        channel=None,
    ):
        model = model or utils.SampleNet(utils.TwoFC())
        channel = channel or instantiate(SketchChannelConfig())
        round_reducer = instantiate(
            SketchRoundReducerConfig(),
            global_model=model,
            channel=channel,
        )
        return round_reducer

    def test_sketch_creation_error(self):
        """
        Tests if creating SketchRoundReducer with no channel raises an error
        """
        model = utils.SampleNet(utils.TwoFC())
        with self.assertRaises((HydraException, AssertionError)):
            _ = instantiate(
                SketchRoundReducerConfig(),
                global_model=model,
                channel=None,
            )

    def test_sketch_creation(self):
        """
        Tests creating SketchRoundReducer correctly
        """
        rr = self.get_round_reducer()
        self.assertIsInstance(rr, SketchRoundReducer)

    def test_sketch_reset(self):
        """
        Tests if reducer's internals resets correctly.
        """
        rr = self.get_round_reducer()
        self.assertTrue(torch.allclose(rr.round_cs.buckets, torch.tensor([0.0])))
        model = utils.SampleNet(utils.TwoFC())

        total_params = 0
        for _, param in model.fl_get_module().named_parameters():
            total_params += param.numel()

        self.assertEqual(rr.round_cs.n, total_params)

        rr.collect_update(model, 1.0)
        rr.reset(model)
        self.assertTrue(torch.allclose(rr.round_cs.buckets, torch.tensor([0.0])))

        self.assertEqual(rr.round_cs.n, total_params)

        for (cs_name, cs_param), (model_name, model_param) in zip(
            list(rr.round_cs.param_sizes.items()),
            list(model.fl_get_module().named_parameters()),
        ):
            self.assertEqual(cs_name, model_name)
            self.assertEqual(cs_param, model_param.size())

    def test_sketch_receive_through_channel(self):
        """
        Tests if reducer's receives CountSketch from channel correctly when given a model
        """
        rr = self.get_round_reducer()
        model = utils.SampleNet(utils.TwoFC())
        message = Message(model, weight=1, count_sketch=CountSketch())
        message = rr.channel.client_to_server(message)
        cs = message.count_sketch

        for (cs_name, cs_param), (model_name, model_param) in zip(
            list(cs.param_sizes.items()), list(model.fl_get_module().named_parameters())
        ):
            self.assertEqual(cs_name, model_name)
            self.assertEqual(cs_param, model_param.size())

    def test_sketch_collect_update_reduce(self):
        """
        Tests if reducer collects update correctly and reduces to the correct weights.
        """
        param_values = [0.1 * i for i in range(100)]
        weights = [i % 10 for i in range(100)]
        global_param = 1.0
        clients = [
            self._fake_client(global_param, p, w) for p, w in zip(param_values, weights)
        ]
        ref_model = utils.SampleNet(utils.TwoFC())
        rr = self.get_round_reducer(ref_model)
        for clnt in clients:
            model, weight = clnt.generate_local_update(utils.SampleNet(utils.TwoFC()))
            rr.collect_update(model, weight)
        expected_sum_weights = sum(weights)
        expected_param_values = sum(
            (global_param - param_value) * w
            for param_value, w in zip(param_values, weights)
        )
        self.assertAlmostEqual(expected_sum_weights, rr.sum_weights.item(), 5)

        sd = rr.round_cs.unsketch_model()
        for _, param in sd.items():
            self.assertTrue(
                torch.allclose(param, torch.tensor([expected_param_values]))
            )

        sketch, sum_weights = rr.reduce()
        self.assertAlmostEqual(expected_sum_weights, sum_weights, 5)
        sd = sketch.unsketch_model()
        for _, param in sd.items():
            self.assertTrue(
                torch.allclose(
                    param, torch.tensor([expected_param_values / expected_sum_weights])
                )
            )



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from __future__ import annotations

import abc
from dataclasses import dataclass
from typing import OrderedDict

from flsim.channels.communication_stats import ChannelStatsCollector
from flsim.channels.message import Message
from flsim.utils.config_utils import fullclassname
from flsim.utils.config_utils import init_self_cfg


class IFLChannel(abc.ABC):
    """
    Base interface for `IFLChannel` that takes care of transmitting messages
    between clients and the server and collecting some metrics on the way.

    This is by nature a *bi-directional* channel (server to client and
    client to server).

    """

    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=FLChannelConfig,
            **kwargs,
        )
        self.stats_collector = (
            ChannelStatsCollector() if self.cfg.report_communication_metrics else None
        )

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    @abc.abstractmethod
    def server_to_client(self, message: Message) -> Message:
        """
        Simulates the manipulation and transmission of a `Message` from
        the server to a client. Also handles relevant stats accounting.
        """
        pass

    @abc.abstractmethod
    def client_to_server(self, message: Message) -> Message:
        """
        Simulates the manipulation and transmission of a `Message` from
        a client to the server. Also handles relevant stats accounting.
        """

        pass


class IdentityChannel(IFLChannel):
    """
    Implements a *bi-directional* channel which is pass-through: the message sent
    is identical to the message received when sending from the server to a
    client and when sending from a client to the server.

    It simulates what happens in the real world. For instance, with the
    direction client to server, there are three methods:
        - ``_on_client_before_transmission``: this is what happens in the client
          right before sending the message, such as compression.
        - ``_during_transmission_client_to_server``: this is what happens during the
          transmission of the message, such as communication measurements,
          communication noise.
        - ``_on_server_after_reception``: this is what happens first when the server
          receives the message, such as decompression.

    Notes:
        - We recommend that all channels inherit `IdentityChannel` and override
          only the necessary methods.
        - Each new channel is responsible for the measurement of its message size,
          i.e. it is the user's responsibility to override the right methods.
    """

    # defining useful general constants for channel size measurements
    BYTES_PER_FP32 = 4
    BYTES_PER_FP64 = 8
    BYTES_PER_INT64 = 8
    BYTES_PER_BIT = 0.125

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    @classmethod
    def calc_model_size_float_point(cls, state_dict: OrderedDict):
        """
        Calculates model size in bytes given a state dict.
        """
        model_size_bytes = sum(
            p.numel() * p.element_size() for (_, p) in state_dict.items()
        )
        return model_size_bytes

    def _calc_message_size_client_to_server(self, message: Message):
        return self.calc_model_size_float_point(message.model_state_dict)

    def _calc_message_size_server_to_client(self, message: Message):
        return self.calc_model_size_float_point(message.model_state_dict)

    def _on_client_before_transmission(self, message: Message) -> Message:
        """
        Implements message manipulation that would be done on a client in the real world
        just before transmitting a message to the server.
        """
        return message

    def _during_transmission_client_to_server(self, message: Message) -> Message:
        """
        Manipulation to the message in transit from client to server.
        """

        if self.stats_collector:
            message_size_bytes = self._calc_message_size_client_to_server(message)
            self.stats_collector.collect_channel_stats(
                message_size_bytes, client_to_server=True
            )
        return message

    def _on_server_after_reception(self, message: Message) -> Message:
        """
        Implements message manipulation that would be done on the server in the real world
        just after receiving a message from a client.
        """
        return message

    def _on_server_before_transmission(self, message: Message) -> Message:
        """
        Implements message manipulation that would be done on the server in the real world
        just before transmitting a message to a client.
        """
        return message

    def _during_transmission_server_to_client(self, message: Message) -> Message:
        """
        Manipulation to the message in transit from server to client.
        """
        if self.stats_collector:
            message_size_bytes = self._calc_message_size_server_to_client(message)
            self.stats_collector.collect_channel_stats(
                message_size_bytes, client_to_server=False
            )
        return message

    def _on_client_after_reception(self, message: Message) -> Message:
        """
        Implements message manipulation that would be done on a client in the real world
        just after receiving a message from the server.
        """
        return message

    def client_to_server(self, message: Message) -> Message:
        """
        Performs three successive steps to send a message from a client to the server.
        """
        # process through channel
        message = self._on_client_before_transmission(message)
        message = self._during_transmission_client_to_server(message)
        message = self._on_server_after_reception(message)

        return message

    def server_to_client(self, message: Message) -> Message:
        """
        Performs three successive steps to send a message from the server to a client.
        """

        # process through channel
        message = self._on_server_before_transmission(message)
        message = self._during_transmission_server_to_client(message)
        message = self._on_client_after_reception(message)

        return message


@dataclass
class FLChannelConfig:
    """
    Base Config for a channel defining channel properties
    such as channel drop rate, quantization effects,
    channel bandwidth settings, etc.
    """

    # Add attributes as needed.
    _target_: str = fullclassname(IdentityChannel)
    _recursive_: bool = False
    # Whether communication metrics (between server and clients) should be
    # reported
    report_communication_metrics: bool = False



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from __future__ import annotations

import abc
from dataclasses import dataclass

from flsim.utils.config_utils import fullclassname
from flsim.utils.config_utils import init_self_cfg
from omegaconf import MISSING


class StalenessWeight(abc.ABC):
    def __init__(self, **kwargs):
        """avg_staleness is used to 'normalize' the weight, such that
        weight=1 when staleness=avg_staleness
        """
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=AsyncStalenessWeightConfig,
            **kwargs,
        )

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    @abc.abstractmethod
    def _raw_weight(self, staleness: int) -> float:
        pass

    def weight(self, staleness: int) -> float:
        assert staleness >= 0, "Staleness must be non-negative"
        # pyre-fixme[16]: `StalenessWeight` has no attribute `cfg`.
        return self._raw_weight(staleness) / self._raw_weight(self.cfg.avg_staleness)


class ConstantStalenessWeight(StalenessWeight):
    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=ConstantStalenessWeightConfig,
            **kwargs,
        )
        super().__init__(**kwargs)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def _raw_weight(self, staleness: int) -> float:
        return 1.0


class ThresholdStalenessWeight(StalenessWeight):
    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=ThresholdStalenessWeightConfig,
            **kwargs,
        )
        super().__init__(**kwargs)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def _raw_weight(self, staleness: int) -> float:
        # pyre-fixme[16]: `ThresholdStalenessWeight` has no attribute `cfg`.
        if staleness <= self.cfg.cutoff:
            return 1.0
        else:
            return self.cfg.value_after_cutoff


class PolynomialStalenessWeight(StalenessWeight):
    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=PolynomialStalenessWeightConfig,
            **kwargs,
        )
        super().__init__(**kwargs)
        assert (
            self.cfg.exponent <= 1 and self.cfg.exponent >= 0
        ), f"PolynomialExponent must be between 0 and 1, inclusive. Got {self.cfg.exponent}"

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def _raw_weight(self, staleness: int) -> float:
        # pyre-fixme[16]: `PolynomialStalenessWeight` has no attribute `cfg`.
        denom = (1 + staleness) ** self.cfg.exponent
        return 1 / denom


@dataclass
class AsyncStalenessWeightConfig:
    _target_: str = MISSING
    _recursive_: bool = False
    avg_staleness: int = 1


@dataclass
class ConstantStalenessWeightConfig(AsyncStalenessWeightConfig):
    _target_: str = fullclassname(ConstantStalenessWeight)


@dataclass
class ThresholdStalenessWeightConfig(AsyncStalenessWeightConfig):
    _target_: str = fullclassname(ThresholdStalenessWeight)
    cutoff: int = MISSING
    value_after_cutoff: float = MISSING


@dataclass
class PolynomialStalenessWeightConfig(AsyncStalenessWeightConfig):
    _target_: str = fullclassname(PolynomialStalenessWeight)
    exponent: float = MISSING



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from __future__ import annotations

import abc
import copy
import logging
from dataclasses import dataclass
from typing import Optional

import torch
from flsim.channels.base_channel import IdentityChannel, IFLChannel
from flsim.common.logger import Logger
from flsim.interfaces.model import IFLModel
from flsim.optimizers.layerwise_optimizers import LAMB, LARS
from flsim.reducers.base_round_reducer import (
    IFLRoundReducer,
    IFLRoundReducerConfig,
)
from flsim.reducers.base_round_reducer import RoundReducerConfig
from flsim.reducers.dp_round_reducer import DPRoundReducerConfig
from flsim.utils.config_utils import fullclassname
from flsim.utils.config_utils import init_self_cfg
from flsim.utils.fl.common import FLModelParamUtils
from hydra.utils import instantiate
from omegaconf import MISSING
from omegaconf import OmegaConf, DictConfig
from torch.nn import Module as Model  # @manual


class SyncAggregator(abc.ABC):
    """
    FL global optimizer for trainers with locally aggregated model
    """

    logger: logging.Logger = Logger.get_logger("SyncAggregator")

    def __init__(
        self,
        *,
        global_model: IFLModel,
        channel: Optional[IFLChannel] = None,
        **kwargs,
    ):
        init_self_cfg(
            self,
            component_class=__class__,  # pyre-fixme[10]: Name `__class__` is used but not defined.
            config_class=SyncAggregatorConfig,
            **kwargs,
        )
        assert (  # TODO: remove after T84246363
            not self.is_round_reducer_dp
        ), "To create a private round reducer, use PrivateSyncTrainer instead."
        self.reducer = instantiate(
            # pyre-fixme[16]: `SyncAggregator` has no attribute `cfg`.
            self.cfg.reducer,
            global_model=global_model,
            channel=channel,
            # TODO: ngjhn passing this here for DP reducer instantiation
            # will fix in T80906721
            num_users_per_round=self.cfg.num_users_per_round,
            total_number_of_users=self.cfg.total_number_of_users,
        )
        self._global_model: IFLModel = global_model

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        if OmegaConf.is_missing(cfg.reducer, "_target_"):
            cfg.reducer = RoundReducerConfig()

    @property
    def is_round_reducer_dp(self):
        # TODO: remove after T84246363
        # reducer can be a DictConfig (if constructed the normal way via constructor)
        # or a dataclass instance (if the param is set directly - not a recommended way).
        return issubclass(self.cfg.reducer.__class__, DPRoundReducerConfig) or (
            isinstance(self.cfg.reducer, DictConfig)
            and issubclass(OmegaConf.get_type(self.cfg.reducer), DPRoundReducerConfig)
        )

    @property
    def global_model(self) -> IFLModel:
        return self._global_model

    def collect_client_update(self, update: IFLModel, weight: float) -> None:
        """
        Collects update from one client and aggregates it internally.
        """
        self.reducer.collect_update(delta=update, weight=weight)

    def init_round(self, reducer: Optional[IFLRoundReducer] = None):
        """
        Just like an optimizer that requires zero_grad to be called
        beginning of each step, FL aggregator requires this function
        to be called beginning of each FL round.
        """
        if reducer is not None and reducer is not self.reducer:
            self.logger.warning("Changing the round reducer!")
            del self.reducer
            self.reducer = reducer
        self.reducer.reset(ref_model=self._global_model)

    @abc.abstractmethod
    def step(self) -> Optional[float]:
        pass


class FedAvgSyncAggregator(SyncAggregator):
    """
    Implements federated averaging
    """

    def __init__(
        self,
        *,
        global_model: IFLModel,
        channel: Optional[IdentityChannel] = None,
        **kwargs,
    ):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=FedAvgSyncAggregatorConfig,
            **kwargs,
        )

        super().__init__(global_model=global_model, channel=channel, **kwargs)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def step(self) -> Optional[float]:
        reduced_module, sum_weights = self.reducer.reduce()
        if not self.reducer.is_averaged:
            raise AttributeError(
                "Reduction type of a FedAvg reducer should be either "
                "AVERAGE or WEIGHTED_AVERAGE."
            )
        if sum_weights > 0:
            FLModelParamUtils.subtract_model(
                minuend=self._global_model.fl_get_module(),
                subtrahend=reduced_module,
                difference=self._global_model.fl_get_module(),
                only_federated_params=True,
            )


class SyncAggregatorWithOptimizer(SyncAggregator):
    """
    Base class for SyncAggregators that use a PyTorch optimizer underneath,
    like FedAvgWithLR and FedAdam
    """

    def __init__(
        self,
        *,
        global_model: IFLModel,
        channel: Optional[IdentityChannel] = None,
        **kwargs,
    ) -> None:
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=SyncAggregatorConfig,
            **kwargs,
        )

        super().__init__(global_model=global_model, channel=channel, **kwargs)
        self.optimizer = create_optimizer_for_sync_aggregator(
            # pyre-fixme[16]: `SyncAggregatorWithOptimizer` has no attribute `cfg`.
            self.cfg,
            global_model.fl_get_module(),
        )
        # creating two temporary models so we don't have to initialize them
        # every time step() is called
        self._reconstructed_grad = copy.deepcopy(self._global_model)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def init_round(self, reducer: Optional[IFLRoundReducer] = None):
        super().init_round(reducer)
        self.optimizer.zero_grad()

    def step(self) -> Optional[float]:
        """
        computes grad and takes an optimizer step:
        grad := (global_model - new_model)
        new_model := global_model - f(grad)
        f() depends on the optimizer:
        e.g. Adam uses first and second moments, LARS and LAMB
        normalize the gradient, etc.
        """
        reduced_module, sum_weights = self.reducer.reduce()
        if sum_weights > 0:
            FLModelParamUtils.set_gradient(
                model=self._global_model.fl_get_module(),
                reference_gradient=reduced_module,
            )
            self.optimizer.step()


class FedAvgWithLRSyncAggregator(SyncAggregatorWithOptimizer):
    def __init__(
        self,
        *,
        global_model: IFLModel,
        channel: Optional[IdentityChannel] = None,
        **kwargs,
    ) -> None:
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=FedAvgWithLRSyncAggregatorConfig,
            **kwargs,
        )

        super().__init__(global_model=global_model, channel=channel, **kwargs)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass


class FedAdamSyncAggregator(SyncAggregatorWithOptimizer):
    def __init__(
        self,
        *,
        global_model: IFLModel,
        channel: Optional[IdentityChannel] = None,
        **kwargs,
    ) -> None:
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=FedAdamSyncAggregatorConfig,
            **kwargs,
        )

        super().__init__(global_model=global_model, channel=channel, **kwargs)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass


class FedLARSSyncAggregator(SyncAggregatorWithOptimizer):
    def __init__(
        self,
        *,
        global_model: IFLModel,
        channel: Optional[IdentityChannel] = None,
        **kwargs,
    ) -> None:
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=FedLARSSyncAggregatorConfig,
            **kwargs,
        )

        super().__init__(global_model=global_model, channel=channel, **kwargs)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass


class FedLAMBSyncAggregator(SyncAggregatorWithOptimizer):
    def __init__(
        self,
        *,
        global_model: IFLModel,
        channel: Optional[IdentityChannel] = None,
        **kwargs,
    ) -> None:
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=FedLAMBSyncAggregatorConfig,
            **kwargs,
        )

        super().__init__(global_model=global_model, channel=channel, **kwargs)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass


def create_optimizer_for_sync_aggregator(config: SyncAggregatorConfig, model: Model):
    if config._target_ == FedAvgWithLRSyncAggregatorConfig._target_:
        return torch.optim.SGD(
            model.parameters(),
            # pyre-fixme[16]: `SyncAggregatorConfig` has no attribute `lr`.
            lr=config.lr,
            # pyre-fixme[16]: `SyncAggregatorConfig` has no attribute `momentum`.
            momentum=config.momentum,
        )
    elif config._target_ == FedAdamSyncAggregatorConfig._target_:
        return torch.optim.Adam(
            model.parameters(),
            lr=config.lr,
            # pyre-fixme[16]: `SyncAggregatorConfig` has no attribute `weight_decay`.
            weight_decay=config.weight_decay,
            # pyre-fixme[16]: `SyncAggregatorConfig` has no attribute `beta1`.
            # pyre-fixme[16]: `SyncAggregatorConfig` has no attribute `beta2`.
            betas=(config.beta1, config.beta2),
            # pyre-fixme[16]: `SyncAggregatorConfig` has no attribute `eps`.
            eps=config.eps,
        )
    elif config._target_ == FedLARSSyncAggregatorConfig._target_:
        return LARS(
            model.parameters(),
            lr=config.lr,
            # pyre-fixme[16]: `SyncAggregatorConfig` has no attribute `beta`.
            beta=config.beta,
            weight_decay=config.weight_decay,
        )

    elif config._target_ == FedLAMBSyncAggregatorConfig._target_:
        return LAMB(
            model.parameters(),
            lr=config.lr,
            beta1=config.beta1,
            beta2=config.beta2,
            weight_decay=config.weight_decay,
            eps=config.eps,
        )


@dataclass
class SyncAggregatorConfig:
    _target_: str = MISSING
    _recursive_: bool = False
    reducer: IFLRoundReducerConfig = IFLRoundReducerConfig()
    # TODO: ngjhn passing this here for DP reducer instantiation
    # will fix in T80906721
    num_users_per_round: int = 1
    total_number_of_users: int = 10000000000


@dataclass
class FedAvgSyncAggregatorConfig(SyncAggregatorConfig):
    _target_: str = fullclassname(FedAvgSyncAggregator)


@dataclass
class FedAvgWithLRSyncAggregatorConfig(SyncAggregatorConfig):
    _target_: str = fullclassname(FedAvgWithLRSyncAggregator)
    lr: float = 0.001
    momentum: float = 0.0


@dataclass
class FedAdamSyncAggregatorConfig(SyncAggregatorConfig):
    _target_: str = fullclassname(FedAdamSyncAggregator)
    lr: float = 0.001
    weight_decay: float = 0.00001
    beta1: float = 0.9
    beta2: float = 0.999
    eps: float = 1e-8


@dataclass
class FedLARSSyncAggregatorConfig(SyncAggregatorConfig):
    _target_: str = fullclassname(FedLARSSyncAggregator)
    lr: float = 0.001
    weight_decay: float = 0.00001
    beta: float = 0.9


@dataclass
class FedLAMBSyncAggregatorConfig(SyncAggregatorConfig):
    _target_: str = fullclassname(FedLAMBSyncAggregator)
    lr: float = 0.001
    weight_decay: float = 0.00001
    beta1: float = 0.9
    beta2: float = 0.999
    eps: float = 1e-8



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from __future__ import annotations

import abc
import copy
from dataclasses import dataclass, field
from typing import List

import torch
from flsim.utils.config_utils import fullclassname
from flsim.utils.config_utils import init_self_cfg
from omegaconf import MISSING
from torch.distributions.exponential import Exponential
from torch.distributions.half_normal import HalfNormal
from torch.distributions.normal import Normal
from torch.distributions.uniform import Uniform


@dataclass
class DurationInfo:
    duration: float = 0


class IDurationDistribution(abc.ABC):
    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=DurationDistributionConfig,
            **kwargs,
        )
        validate_args = False if self.cfg.training_duration_sd == 0 else True
        self.gaussian_generator: Normal = Normal(
            torch.tensor([self.cfg.training_duration_mean], dtype=torch.float),
            torch.tensor([self.cfg.training_duration_sd], dtype=torch.float),
            validate_args=validate_args,
        )

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def bounded_gaussian_sample(self) -> float:
        raw_sample = self.gaussian_generator.sample().item()
        # pyre-fixme[16]: `IDurationDistribution` has no attribute `cfg`.
        return max(raw_sample, self.cfg.training_duration_min)

    @abc.abstractmethod
    def training_duration(self, num_training_examples: int) -> float:
        pass


class PerExampleGaussianDurationDistribution(IDurationDistribution):
    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=PerExampleGaussianDurationDistributionConfig,
            **kwargs,
        )
        super().__init__(**kwargs)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def training_duration(self, num_training_examples: int) -> float:
        one_example_duration = self.bounded_gaussian_sample()
        return num_training_examples * one_example_duration


class PerUserGaussianDurationDistribution(IDurationDistribution):
    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=PerUserGaussianDurationDistributionConfig,
            **kwargs,
        )
        super().__init__(**kwargs)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def training_duration(self, num_training_examples: int) -> float:
        return self.bounded_gaussian_sample()


class PerUserHalfNormalDurationDistribution(IDurationDistribution):
    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=PerUserHalfNormalDurationDistributionConfig,
            **kwargs,
        )
        super().__init__(**kwargs)

        self.generator: HalfNormal = HalfNormal(scale=self.cfg.training_duration_sd)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def training_duration(self, num_training_examples: int) -> float:
        # pyre-fixme[16]: `PerUserHalfNormalDurationDistribution` has no attribute
        #  `cfg`.
        return self.cfg.training_duration_min + self.generator.sample()


class PerUserUniformDurationDistribution(IDurationDistribution):
    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=PerUserUniformDurationDistributionConfig,
            **kwargs,
        )
        super().__init__(**kwargs)

        assert (
            self.cfg.training_duration_sd == 0.0
        ), "Cannot set training duration sd for uniform distribution"
        self.generator: Uniform = Uniform(
            low=self.cfg.training_duration_min,
            high=2 * self.cfg.training_duration_mean - self.cfg.training_duration_min,
        )

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def training_duration(self, num_training_examples: int) -> float:
        # pyre-ignore[20]: default return shape is 1
        return self.generator.sample().item()


class PerUserExponentialDurationDistribution(IDurationDistribution):
    """
    Exponetial Duration where training_duration_mean is the rate parameter
    """

    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=PerUserExponentialDurationDistributionConfig,
            **kwargs,
        )
        super().__init__(**kwargs)

        assert (
            self.cfg.training_duration_sd == 0.0
        ), "Cannot set training duration sd for exponetial"
        self.generator: Exponential = Exponential(
            rate=1 / self.cfg.training_duration_mean
        )

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def training_duration(self, num_training_examples: int) -> float:
        # pyre-ignore[20]: default return shape is 1
        return self.generator.sample().item()


class DurationDistributionFromList(IDurationDistribution):
    """
    This class simulates IDurationDistribution
    It returns traing duration from a fixed list
    Useful for writing unit tests for components that use TrainingEventGenerator
    """

    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=DurationDistributionFromListConfig,
            **kwargs,
        )
        super().__init__(**kwargs)
        self.training_events = list(self.cfg.training_events)

        self.distr: List[DurationInfo] = copy.deepcopy(self.training_events)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def training_duration(self, num_training_examples: int) -> float:
        return self.distr.pop(0).duration


@dataclass
class DurationDistributionConfig:
    _target_: str = MISSING
    _recursive_: bool = False
    training_duration_mean: float = 0.0
    training_duration_sd: float = 0.0
    training_duration_min: float = float("-inf")


@dataclass
class PerExampleGaussianDurationDistributionConfig(DurationDistributionConfig):
    _target_: str = fullclassname(PerExampleGaussianDurationDistribution)


@dataclass
class PerUserGaussianDurationDistributionConfig(DurationDistributionConfig):
    _target_: str = fullclassname(PerUserGaussianDurationDistribution)


@dataclass
class PerUserHalfNormalDurationDistributionConfig(DurationDistributionConfig):
    _target_: str = fullclassname(PerUserHalfNormalDurationDistribution)


@dataclass
class PerUserUniformDurationDistributionConfig(DurationDistributionConfig):
    _target_: str = fullclassname(PerUserUniformDurationDistribution)


@dataclass
class PerUserExponentialDurationDistributionConfig(DurationDistributionConfig):
    _target_: str = fullclassname(PerUserExponentialDurationDistribution)


@dataclass
class DurationDistributionFromListConfig(DurationDistributionConfig):
    _target_: str = fullclassname(DurationDistributionFromList)
    training_events: List[DurationInfo] = field(default_factory=list)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from flsim.utils.fl.common import FLModelParamUtils
from flsim.utils.fl.personalized_model import FLModelWithPrivateModules
from flsim.utils.tests.helpers.test_models import (
    FC_PRIVATE_MODULE_NAMES,
    PersonalizedFCModel,
)
from libfb.py import testutil


class FLModelWithPrivateModulesTest(testutil.BaseFacebookTestCase):
    def setUp(self) -> None:
        super().setUp()
        self.model_with_private_modules = PersonalizedFCModel()

    def test_clear_dict_module(self) -> None:
        # number of private parameters are number of fc layers in
        # private_module_names times 2 (W and b in each layer)
        self.assertEqual(
            len(list(FLModelWithPrivateModules.get_user_private_parameters())),
            len(FC_PRIVATE_MODULE_NAMES) * 2,
        )
        FLModelWithPrivateModules.clear_user_private_module_dict()
        self.assertEmpty(list(FLModelWithPrivateModules.get_user_private_parameters()))

    def test_get_user_private_attr(self) -> None:
        fc1_layer = self.model_with_private_modules.get_user_private_attr("fc1")
        self.assertEqual(
            FLModelParamUtils.get_num_trainable_params(fc1_layer), 10 * 5 + 5
        )
        with self.assertRaises(AttributeError):
            self.model_with_private_modules.get_user_private_attr("fc2")



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

"""
This file contains the noise generation function and the required
DP parameters that an entity such as an FL server uses for user-level DP.
"""

import logging
import math
import os
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional, List

import numpy as np
import torch
from flsim.common.logger import Logger
from flsim.privacy.common import (
    PrivacyBudget,
    PrivacySetting,
)
from opacus import privacy_analysis
from torch import nn


class PrivacyEngineNotAttachedException(Exception):
    """
    Exception class to be thrown from User Privacy Engine in case
    the User Privacy Engine is not attached.
    """

    pass


class IPrivacyEngine(ABC):
    def __init__(
        self,
        privacy_setting: PrivacySetting,
        users_per_round: int,
        num_total_users: int,
    ):
        self.setting = privacy_setting
        self.users_per_round = users_per_round
        self.num_total_users = num_total_users
        self.steps = 0

    @abstractmethod
    def attach(self, global_model: nn.Module) -> None:
        """
        Attach the privacy engine to the global model by setting
        a reference model
        """
        pass

    @abstractmethod
    def add_noise(self, model_diff: nn.Module, sensitivity: float) -> None:
        pass

    @abstractmethod
    def get_privacy_spent(self, target_delta: Optional[float] = None) -> PrivacyBudget:
        pass


class GaussianPrivacyEngine(IPrivacyEngine):
    """
    DP-SGD privacy engine where noise is independent
    and comes from a gaussian distribution
    """

    logger: logging.Logger = Logger.get_logger(__name__)

    def __init__(
        self,
        privacy_setting: PrivacySetting,
        users_per_round: int,
        num_total_users: int,
    ) -> None:

        super().__init__(privacy_setting, users_per_round, num_total_users)
        self.noise_multiplier = privacy_setting.noise_multiplier
        self.target_delta = privacy_setting.target_delta
        self.alphas = privacy_setting.alphas
        self.user_sampling_rate = float(users_per_round) / num_total_users
        self.device = None
        self.random_number_generator = None
        self.noise_seed = privacy_setting.noise_seed

    def attach(self, global_model: nn.Module):
        self.device = next(global_model.parameters()).device

        noise_seed = (
            int.from_bytes(os.urandom(8), byteorder="big", signed=True)
            if self.noise_seed is None
            else self.noise_seed
        )
        torch.cuda.manual_seed_all(noise_seed)  # safe to call even if no gpu.
        self.random_number_generator = torch.Generator(  # pyre-ignore
            device=self.device
        )
        # pyre-fixme[16]: `Generator` has no attribute `manual_seed`.
        self.random_number_generator.manual_seed(noise_seed)
        self.logger.debug("User privacy engine attached.")

    def add_noise(self, model_diff: nn.Module, sensitivity: float) -> None:
        """
        Adds noise to the model_diff (operation is in-place).

        This method adds noise to the parameters of the input model.
        This operation is in-place (modifies model_diff in this method)
        Noise is sampled from a normal distribution with 0 mean and
        standard deviation equal to sensitivity * noise_multiplier.

        Parameters
        ----------
        model_diff : nn.Module
            Noise will be added to the parameters of this model.
        sensitivity : float
            The sensitivity of the noise that will be added.
        """
        with torch.no_grad():
            for _, parameter in model_diff.named_parameters():
                noise = self._generate_noise(parameter.shape, sensitivity)
                parameter.copy_(parameter + noise)
            self.steps += 1

    def _generate_noise(self, size, sensitivity: float) -> torch.Tensor:
        if self.device is None or self.random_number_generator is None:
            random_gen = "no" if self.random_number_generator is None else "yes"
            raise PrivacyEngineNotAttachedException(
                "User Privacy Engine is not attached to the global model. "
                f"(device={self.device}, random number generator exists: {random_gen})."
                "Call attach() function first before calling."
            )
        if self.noise_multiplier > 0 and sensitivity > 0:
            return torch.normal(
                0,
                self.noise_multiplier * sensitivity,
                size,
                device=self.device,
                generator=self.random_number_generator,
            )
        return torch.zeros(size, device=self.device)

    def get_privacy_spent(self, target_delta: Optional[float] = None):
        if target_delta is None:
            target_delta = self.target_delta

        rdp = privacy_analysis.compute_rdp(
            self.user_sampling_rate, self.noise_multiplier, self.steps, self.alphas
        )
        eps, opt_alpha = privacy_analysis.get_privacy_spent(
            self.alphas, rdp, delta=target_delta
        )

        self.logger.info(
            f"User-level DP Privacy Parameters:"
            f"\tuser sampling rate = {100 * self.user_sampling_rate:.3g}%"
            f"\tnoise_multiplier = {self.noise_multiplier}"
            f"\tsteps = {self.steps}\n  satisfies "
            f"DP with Ɛ = {eps:.3g} "
            f"and δ = {target_delta}."
            f"  The optimal α is {opt_alpha}."
        )
        if opt_alpha == max(self.alphas) or opt_alpha == min(self.alphas):
            self.logger.info(
                "The privacy estimate is likely to be improved by expanding "
                "the set of alpha orders."
            )
        return PrivacyBudget(eps, opt_alpha, target_delta)


@dataclass
class TreeNode:
    start: int
    end: int
    height: int
    efficient: bool

    @property
    def weight(self):
        return (1 / (2 - math.pow(2, -self.height))) ** 0.5 if self.efficient else 1.0


class TreePrivacyEngine(IPrivacyEngine):
    """
    DP-FTRL privacy engine where noise is the cummulated noise from
    a private binary tree
    """

    logger: logging.Logger = Logger.get_logger(__name__)

    def __init__(
        self,
        privacy_setting: PrivacySetting,
        users_per_round: int,
        num_total_users: int,
        restart_rounds: int = 100,
        efficient_tree: bool = True,
    ) -> None:
        super().__init__(privacy_setting, users_per_round, num_total_users)
        self.num_leaf = min(num_total_users, restart_rounds * users_per_round)
        self.restart_rounds = restart_rounds
        self.num_tree: int = 1
        self.ref_model = None
        self.device = None

        self.tree = TreePrivacyEngine.build_tree(self.num_leaf, efficient_tree)

    @classmethod
    def build_tree(cls, num_leaf: int, efficient_tree: bool = True) -> List[TreeNode]:
        tree = [TreeNode(-1, -1, -1, True)] * (2 * num_leaf)
        # store leaf nodes at back of array
        for i, j in enumerate(range(num_leaf, num_leaf * 2)):
            tree[j] = TreeNode(start=i, end=i, height=0, efficient=efficient_tree)

        # fill in prefix sum internal nodes
        for i in range(num_leaf - 1, 0, -1):
            left = tree[i * 2]
            right = tree[i * 2 + 1]
            height = int(math.log2(abs(right.end - left.start) + 1))
            tree[i] = TreeNode(
                start=left.start, end=right.end, height=height, efficient=efficient_tree
            )
        return tree

    @classmethod
    def compute_rdp(cls, alphas, epoch, steps, sigma):
        alphas = np.array(alphas)
        return alphas * epoch * np.ceil(np.log2(steps + 1e-6)) / (2 * sigma ** 2)

    def get_privacy_spent(self, target_delta: Optional[float] = None) -> PrivacyBudget:
        target_delta = (
            self.setting.target_delta if target_delta is None else target_delta
        )

        rdp = TreePrivacyEngine.compute_rdp(
            alphas=self.setting.alphas,
            epoch=self.num_tree,
            steps=self.num_leaf,
            sigma=self.setting.noise_multiplier,
        )
        eps, opt_alpha = privacy_analysis.get_privacy_spent(
            orders=self.setting.alphas, rdp=rdp, delta=target_delta
        )

        if opt_alpha == max(self.setting.alphas) or opt_alpha == min(
            self.setting.alphas
        ):
            self.logger.info(
                "The privacy estimate is likely to be improved by expanding "
                "the set of alpha orders."
            )
        return PrivacyBudget(eps, opt_alpha, self.setting.target_delta)

    def attach(self, global_model: nn.Module, **kwargs) -> None:
        """
        Reset the tree by incrementing num_tree and reset steps to 0
        these will be used to do privacy calculations
        """
        self.device = next(global_model.parameters()).device
        self.num_tree += 1
        self.steps = 0

    def add_noise(self, model_diff: nn.Module, sensitivity: float) -> None:
        """
        Adds noise to cummulated noise to model diff
        """
        with torch.no_grad():
            for parameter in model_diff.parameters():
                noise = self.range_sum(
                    left=0,
                    right=self.users_per_round - 1,
                    size=parameter.shape,
                    sensitivity=sensitivity,
                )
                parameter.copy_(parameter + noise)

            self.steps += 1

    def range_sum(
        self, left: int, right: int, size: torch.Size, sensitivity: float
    ) -> torch.Tensor:
        left += self.num_leaf
        right += self.num_leaf
        sum_ = torch.zeros(size)
        while left <= right:
            noise_std = self.setting.noise_multiplier * sensitivity

            if left % 2 == 1:
                sum_ += self._generate_noise(size, noise_std) * self.tree[left].weight

                left += 1

            if right % 2 == 0:
                sum_ += self._generate_noise(size, noise_std) * self.tree[right].weight
                right -= 1

            left = left // 2
            right = right // 2
        return sum_

    def _generate_noise(self, size: torch.Size, noise_std: float):
        return torch.normal(
            mean=0,
            std=noise_std,
            size=size,
            device=self.device,
        )



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import logging
from enum import IntEnum
from itertools import chain
from typing import List, Iterable, Tuple

import torch
import torch.distributed as dist
import torch.multiprocessing as mp
import torch.nn as nn
from flsim.common.logger import Logger
from flsim.utils.fl.common import FLModelParamUtils


class OperationType(IntEnum):
    BROADCAST = 0
    SUM_AND_BROADCAST = 1
    SUM = 2


class FLDistributedUtils:
    """
    More detailed note: https://fburl.com/5n8mnaf3. We cannot use PyTorch
    DDP here, because DDP is tied to backward() and only provides
    high-level APIs for reducing gradients, but in FL, after each round, we
    need to perform all-reduce on the models (optionally with some pre and
    post-processing) instead of just model gradients (e.g. FedAvg).
    We reduce the number of all-reduce operations by flatten an entire model
    into a 1D tensor, if the model size is less than a buffer limit of 256MB.
    For large models, we group the model parameters into flatten buckets of
    256MB each and call all-reduce() in each bucket using async operations.
    """

    logger: logging.Logger = Logger.get_logger(__name__)

    # equivalent to 256 MB of floats, same buffer size as in PyTorch DDP
    MAX_BUFFER_SIZE = 2 ** 28
    WORLD_SIZE = 1
    NUM_WORKERS = 1
    # allow distributed training on CPU, default False
    DISTRIBUTED_TRAINING_ON_CPU = False
    DISTRIBUTED_BACKED = dist.Backend.NCCL

    @classmethod
    def distributed_training_on_cpu(cls):
        cls.DISTRIBUTED_TRAINING_ON_CPU = True
        cls.DISTRIBUTED_BACKED = dist.Backend.GLOO

    @classmethod
    def distributed_training_on_cuda(cls):
        """
        this is on by default, use if you have called
        distributed_training_on_cpu and want to enable
        cpu distributed again.
        """
        cls.DISTRIBUTED_TRAINING_ON_CPU = False
        cls.DISTRIBUTED_BACKED = (
            dist.Backend.GLOO if cls.WORLD_SIZE > cls.NUM_WORKERS else dist.Backend.NCCL
        )

    @classmethod
    def check_distributed_training_configs(
        cls, distributed_world_size: int, cuda: bool
    ) -> Tuple[int, int]:
        """
        Checks resources on the machine and r
        eturns the distributed world size and the number of workers.

        For cpu we do not allow more than one processes per cpu.
        For cuda we do
        """
        if cuda:
            assert torch.cuda.is_available(), (
                "distributed_world_size is greater than 1 "
                "use only if cuda is supported or distributed_training_on_cuda"
                "has been called!"
            )
            num_gpus = torch.cuda.device_count()
            if distributed_world_size > num_gpus and distributed_world_size % num_gpus:
                raise Warning(
                    f"There are {num_gpus} physical cuda workers (i.e gpus), "
                    f"you are asking {distributed_world_size} workers, "
                    "we need equal number of workers per gpu"
                )  # TODO for now we throw may be later on we just print the warning
            return distributed_world_size // num_gpus * num_gpus, num_gpus
        else:
            num_cpus = mp.cpu_count()
            if distributed_world_size > num_cpus:
                raise Warning(
                    f"Only {num_cpus} CPUs are available, "
                    f"but {distributed_world_size} workers were requested."
                )  # TODO for now we throw may be later on we just print the warning
            return min(distributed_world_size, num_cpus), num_cpus

    @classmethod
    def setup_distributed_training(cls, distributed_world_size: int, use_cuda=True):
        cls.WORLD_SIZE, cls.NUM_WORKERS = cls.check_distributed_training_configs(
            distributed_world_size, use_cuda
        )
        if use_cuda:
            cls.distributed_training_on_cuda()
        else:
            cls.distributed_training_on_cpu()

    @classmethod
    def distributed_operation(
        cls,
        params: Iterable[torch.Tensor],
        op: OperationType = OperationType.SUM_AND_BROADCAST,
        src: int = -1,
        dst: int = -1,
    ):
        """
        Group params into a list of flatten buffers and call the distributed
        operation on each buffer asynchronously.

        The actual async operation for each buffer is done in the helper function
        `_distributed_operation`

        Starting with an unprocessed buffer, loops over params and does one of the following:

        * appends the param to the current unprocessed buffer if buffer has space
        * if buffer cannot fit the param, if the param can fit into a new buffer
          sends the current buffer `_distributed_operation` and creates a new buffer
          or else sends param to `_distributed_operation` and keeps the buffer for the
          next param in the list.

        At the end the function joins all async ops and puts processed values from each flattened
        buffer in to their respective param.

        Note:
            In all operations it is assumed that the master worker is the worker with rank 0.

        """

        if cls.WORLD_SIZE == 1:
            return
        # temp variable of list of model params sent organized into one buffer
        operation_results = []  # operation results a list of (handle, buffer)
        param_references = []  # list of param-lists in each buffer
        buffered_params = []  # buffer to hord tensors until enough for dist operation

        offset = 0
        with torch.no_grad():
            for param in params:
                sz = param.numel()
                if sz + offset <= cls.MAX_BUFFER_SIZE:
                    # append the params and postpone the operation
                    buffered_params.append(param)
                    offset += sz
                    continue
                # do the operation, the buffer cannot be appended anymore
                process_independently = sz > cls.MAX_BUFFER_SIZE
                tensor_list = [param] if process_independently else buffered_params
                operation_result = cls._distributed_operation(
                    tensor_list, sz, op, src, dst
                )  # operation result is a tuple of (handle, buffer)
                operation_results.append(operation_result)
                param_references.append(tensor_list)
                offset = offset if process_independently else sz
                buffered_params = buffered_params if process_independently else [param]

            if len(buffered_params) > 0:
                operation_result = cls._distributed_operation(
                    buffered_params, offset, op, src, dst
                )  # operation result is a tuple of (handle, buffer)
                operation_results.append(operation_result)
                param_references.append(buffered_params)

        # wait on the async handle
        for handle, _ in operation_results:
            handle.wait()

        # copy data from flattened buffers to the actual tensors.
        for params, (_, buffer) in zip(param_references, operation_results):
            cls._get_params_from_buffer(params, buffer)

    @classmethod
    def _distributed_operation(
        cls,
        params: List[torch.Tensor],
        numels: int,
        op: OperationType,
        src: int = -1,
        dst: int = -1,
    ):
        """
        Returns a tuple of handle and buffer. Caller is RESPONSIBLE for awaiting
        on handle and then use whatever that's filled in the buffer.
        Creates a buffer of the size of 'numels'. Then, we loop over the
        'params', which is a list of tensors, and copy each tensor (which is
        avset of parameters from model) to buffer one by one. After that, we
        callvall_reduce() function in PyTorch distributed as an async
        operation tovall processes in the group (and get async handle to
        return after this).

        Args:
            params: List[torch.Tensor], a buffer group of parameters to perform
            async operation at one time
            numels: total number of scalar elements in params

        Returns:
            handle: an async handle
            buffer: within distributed operation, params: List[torch.Tensor] is flattened
            as a buffer (1D Tensor) and sent to all_reduce. buffer will store the
            result of distributed option once it is finished.

        Note:
            Size of each param in params are not required to be the same. params is first flatten
            to a 1D tensor. E.g:

                params = Tensor(
                    [1,2,3,4], [ [5,6], [7,8] ], [9,10]
                )

                then buffer is

                [1,2,3,4,5,6,7,8,9,10]

        Example:
            if worker 1 has
                params = [
                    Tensor([1,2,3,4]),
                    Tensor([ [5,6], [7,8] ]),
                    Tensor([9,10])
                ]

            and worker 2 has
                params = [
                    Tensor([10,20,30,40]),
                    Tensor([ [50,60], [70,80] ]),
                    Tensor([90,100])
                ]

            and if the operation type is sum, the returned buffer will be:

            Tensor([11, 22, 33, 44, 55, 66, 77, 88, 99, 110])

        """
        # TODO: enable all_reduce on mixed dtypes with dtype-based bucketing
        # currently the assumption is that there is at least one float tensor
        # so all layers could be casted to float
        # NOTE: seems to work for mixed int and float types
        generic_type = torch.float
        for p in params:
            if p.dtype != generic_type:
                cls.logger.warning("non float tensor types sent to all reduce")

        buffer = params[0].new_empty(numels, dtype=generic_type)
        offset = 0
        for p in params:
            sz = p.numel()
            buffer[offset : offset + sz].copy_(p.data.view(-1))
            offset += sz

        if op == OperationType.SUM_AND_BROADCAST:
            handle = dist.all_reduce(
                buffer,
                op=dist.ReduceOp.SUM,
                group=cls._get_default_group(),
                async_op=True,
            )
        elif op == OperationType.SUM:
            if dst < 0:
                cls.logger.debug("dst is not defined setting 0 as the default value")
                dst = 0
            cls.logger.warning("Operation reduce is not supported on CPU on ")

            if not (
                cls.DISTRIBUTED_TRAINING_ON_CPU
                or cls.DISTRIBUTED_BACKED == dist.Backend.NCCL
            ):
                # GLOO on GPU does not support reduce
                cls.logger.warning("Chaning reduce operation to reduce all.")
                handle = dist.all_reduce(
                    buffer,
                    op=dist.ReduceOp.SUM,
                    group=cls._get_default_group(),
                    async_op=True,
                )
            else:
                handle = dist.reduce(
                    buffer,
                    dst,
                    op=dist.ReduceOp.SUM,
                    group=cls._get_default_group(),
                    async_op=True,
                )
        elif op == OperationType.BROADCAST:
            if src < 0:
                cls.logger.debug(
                    "Distributed copy operation (broadcast) needs a source."
                    "Assigning 0 as the default source"
                )
                src = 0
            handle = dist.broadcast(
                buffer,
                src,
                group=cls._get_default_group(),
                async_op=True,
            )
        else:
            raise ValueError(f"Operation {op} not found. Please check the parameters.")
        return (handle, buffer)

    @classmethod
    def _get_params_from_buffer(cls, params: List[torch.Tensor], buffer: torch.Tensor):
        """
        Inverse the buffering operation in all_reduce and copies the data
        in buffer into each param in params.
        i.e. Copies all-reduced grads back into their original place. However,
        more generally speaking, what this function actually does is treating the
        'buffer' (i.e. the 2nd param) as a well-flattened 1D tensor of the list
        of params and copy all the params back to buffer.
        """
        # TODO: (jesikmin) T55869097 Check whether the size of buffer is same as
        # the total number of elements of params
        # copy all-reduced grads back into their original place
        offset = 0
        for p in params:
            sz = p.numel()
            p.data.copy_(buffer[offset : offset + sz].view_as(p))
            offset += sz

    @classmethod
    def _get_default_group(cls):
        return dist.group.WORLD

    @classmethod
    def is_master_worker(cls):
        """
        we assume that worker 0 is the master worker."
        """
        return (not dist.is_initialized()) or dist.get_rank() == 0

    @classmethod
    def suppress_output(cls):
        import builtins as __builtin__

        builtin_print = __builtin__.print

        def print(*args, **kwargs):
            # force print the result when kwargs contains force and value is True
            if kwargs.pop("force", False):
                builtin_print(*args, **kwargs)

        __builtin__.print = print

    @classmethod
    def dist_init(
        cls,
        rank: int,
        world_size: int,
        init_method: str,
        use_cuda: bool = True,
    ):
        cls.setup_distributed_training(world_size, use_cuda)
        if not cls.DISTRIBUTED_TRAINING_ON_CPU:
            device = torch.device(f"cuda:{rank % cls.NUM_WORKERS}")
            torch.cuda.set_device(device)
        if world_size > 1:
            dist.init_process_group(
                backend=cls.DISTRIBUTED_BACKED,
                init_method=init_method,
                world_size=world_size,
                rank=rank,
            )

    @classmethod
    def synchronize_across_ranks(
        cls,
        model: nn.Module,
        weights: torch.Tensor,
        operation: OperationType,
        only_federated_params: bool = False,
    ):
        state_dict = FLModelParamUtils.get_state_dict(
            model, only_federated_params=only_federated_params
        )
        cls.distributed_operation(
            params=chain([weights], state_dict.values()), op=operation
        )



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

"""Pyre complains about non-default class members in config being not initialized
"""
from __future__ import annotations

import abc
from dataclasses import dataclass
from math import log10, sqrt

from flsim.utils.config_utils import fullclassname
from flsim.utils.config_utils import init_self_cfg
from omegaconf import MISSING


class ExampleWeight(abc.ABC):
    def __init__(self, **kwargs):
        """avg_num_examples is used to 'normalize' the weight, such that
        weight=1 for the average user
        """
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=AsyncExampleWeightConfig,
            **kwargs,
        )

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    @abc.abstractmethod
    def _raw_weight(self, num_examples: float) -> float:
        pass

    def weight(self, num_examples: float) -> float:
        assert num_examples > 0, "Num examples must be positive"
        return self._raw_weight(num_examples) / self._raw_weight(
            # pyre-fixme[16]: `ExampleWeight` has no attribute `cfg`.
            self.cfg.avg_num_examples
        )


class EqualExampleWeight(ExampleWeight):
    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=EqualExampleWeightConfig,
            **kwargs,
        )

        super().__init__(**kwargs)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def _raw_weight(self, num_examples: float) -> float:
        return 1.0


class LinearExampleWeight(ExampleWeight):
    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=LinearExampleWeightConfig,
            **kwargs,
        )

        super().__init__(**kwargs)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def _raw_weight(self, num_examples: float) -> float:
        return num_examples


class SqrtExampleWeight(ExampleWeight):
    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=SqrtExampleWeightConfig,
            **kwargs,
        )

        super().__init__(**kwargs)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def _raw_weight(self, num_examples: float) -> float:
        return sqrt(num_examples)


class Log10ExampleWeight(ExampleWeight):
    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=Log10ExampleWeightConfig,
            **kwargs,
        )

        super().__init__(**kwargs)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def _raw_weight(self, num_examples: float) -> float:
        return log10(1 + num_examples)


@dataclass
class AsyncExampleWeightConfig:
    _target_: str = MISSING
    _recursive_: bool = False
    avg_num_examples: int = 1


@dataclass
class EqualExampleWeightConfig(AsyncExampleWeightConfig):
    _target_: str = fullclassname(EqualExampleWeight)


@dataclass
class LinearExampleWeightConfig(AsyncExampleWeightConfig):
    _target_: str = fullclassname(LinearExampleWeight)


@dataclass
class SqrtExampleWeightConfig(AsyncExampleWeightConfig):
    _target_: str = fullclassname(SqrtExampleWeight)


@dataclass
class Log10ExampleWeightConfig(AsyncExampleWeightConfig):
    _target_: str = fullclassname(Log10ExampleWeight)
